/*
 * TouCAN - common CANopen/CAL driver part for Motorola TouCAN compatible
 *
 * Copyright (c) 1998,2000 port GmbH Halle (Saale)
 *------------------------------------------------------------------
 * $Header: /z2/cvsroot/library/co_lib/drivers/shar_src/Attic/TouCAN.c,v 1.9 2001/04/19 15:23:01 ro Exp $
 *
 *------------------------------------------------------------------
 *
 * modification history
 * --------------------
 * $Log: TouCAN.c,v $
 * Revision 1.9  2001/04/19 15:23:01  ro
 * adapt to CANopenLIB V4.2
 * Set_Baudrate added
 * Clear_busoff added
 * Temporary solution for Problem with SYNC-TPDO's.
 *
 * Revision 1.8  2001/04/11 11:30:05  ro
 * Interrupt counter changed
 * PRINTF macros
 *
 * Revision 1.7  2000/06/06 08:59:19  ro
 * deutsche Handbuchkommentare ergnzt.
 *
 * Revision 1.6  2000/04/04 08:46:32  boe
 * expanded for multiline version
 *
 * Revision 1.5  2000/02/10 13:26:50  boe
 * Anpassung an Library V4.0
 *
 *
 *
 *------------------------------------------------------------------
 */

/*
* DESCRIPTION
* Version $Revision: 1.9 $
* .sp
*++ This module contains 
*++ all functions needed for a CANopen/CAL CAN driver
*++ which are specific to the Motorola TouCAN architecture.
*
*++ In this Version for using with CANopen/CAL, it uses only
*++ standard identifier format.
*
*-- Dieses Modul beinhaltet alle Funktionen, welche fr den
*-- CANopen/CAL CAN Treiber mit den spezifischen Merkmalen
*-- der Motorola TouCAN Architektur gebraucht werden.
*
*-- In dieser Version bei der Benutzung mit CANopen/CAL 
*-- wird nur das standard identifier format benutzt.
*
*++ In BASIC CAN mode the driver uses only two Message Objects
*++ The 'Last Message Object' to receive all Messages
*++ and a defined Message Object , TRANSMIT_OBJ, to transmit Messages.
*
*-- Im Basic CAN Mode benutzt der Treiber nur zwei Message Objekte.
*-- Das 'Last Message Object' zum Empfangen aller Nachrichten
*-- und ein definiertes Message Object namens TRANSMIT_OBJ zum
*-- senden der Messages.
*
*++ If CONFIG_FULLCAN is defined, all Message Objects of the CAN controller
*++ are used.
*++ Message Objects are allocated in the order they are requested.
*++ If in this mode also CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
*++ is defined, then two predifined transmit channels are used.
*++ One for common transmit requests and one for the Node Guarding channel.
*++ All other channels from 2...14 are available for receiving
*++ and are allocated using
*++ .I Define_COB()
*++ requests for receive objects.
*
*-- Wenn CONFIG_FULLCAN definiert ist, werden alle Messages Objekte
*-- des CAN Controllers benutzt.
*-- Der Message Objekte werden in der Reihenfolge ihrer Anforderung 
*-- zugeordnet. 
*-- Ist in diesem Mode auch CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
*-- definiert, werden zwei vordefinierte bertragungskanle benutzt.
*-- Einer fr allgemeine Transmit Requests und einer fr den 
*-- Node Guarding Channel. Alle anderen Kanle von 2..14 sind fr
*-- den Empfang verfgbar und knnen per
*-- .I Define_COB()
*-- angefordert werden.
*
*/

extern UNSIGNED16	canIntLevel;

# ifdef CONFIG_MULT_LINES
tou_can_t *tou_can_addr[CONFIG_MULT_LINES];
# else
tou_can_t *tou_can;
# endif

volatile UNSIGNED32 cal_ch_ints = 0;            /* DBG */
volatile UNSIGNED32 cal_tx_ints = 0;            /* DBG */
volatile UNSIGNED32 cal_rx_ints = 0;            /* DBG */

/*--- bit timing table --------------------------------------------
 * !!! The TouCAN driver need more parameters then many other
 * !!! drivers
 * !!! you must use BTR_TAB_TOUCAN_T and cast the pointers to the table
 *-------------------------------------------------------------------*/
const BTR_TAB_TOUCAN_T can_btr_tab_toucan[] = {
 { 10, CAN_PRESDIV_10K, CAN_PROPSEG_10K, CAN_PSEG1_10K, CAN_PSEG2_10K },
 { 20, CAN_PRESDIV_20K, CAN_PROPSEG_20K, CAN_PSEG1_20K, CAN_PSEG2_20K },
 { 50, CAN_PRESDIV_50K, CAN_PROPSEG_50K, CAN_PSEG1_50K, CAN_PSEG2_50K },
 { 100, CAN_PRESDIV_100K, CAN_PROPSEG_100K, CAN_PSEG1_100K, CAN_PSEG2_100K },
 { 125, CAN_PRESDIV_125K, CAN_PROPSEG_125K, CAN_PSEG1_125K, CAN_PSEG2_125K },
 { 250, CAN_PRESDIV_250K, CAN_PROPSEG_250K, CAN_PSEG1_250K, CAN_PSEG2_250K },
 { 500, CAN_PRESDIV_500K, CAN_PROPSEG_500K, CAN_PSEG1_500K, CAN_PSEG2_500K },
 { 800, CAN_PRESDIV_800K, CAN_PROPSEG_800K, CAN_PSEG1_800K, CAN_PSEG2_800K },
 { 1000, CAN_PRESDIV_1000K, CAN_PROPSEG_1000K, CAN_PSEG1_1000K, CAN_PSEG2_1000K },
 {0, 0, 0, 0, 0}  /* last entry */
};


/*******************************************************************
*
*++ Init_CAN - initialize the CAN-Controller
*-- Init_CAN - initialisiert den CAN-Controller
*
*++ BaseAdrr - should used, 
*++ the 68376 has more then one adress area,
*++ select with the value from Module Mapping bit (MM) in SIMCR.
*
*-- BaseAddr - sollte benutzt werden, zumindest als Angabe des Bereiches,
*-- da beim 68376 auer dem Standardadressbereich noch ein anderer mglich ist.
*-- Je nach Wert des Module Mapping bit (MM) im  SIMCR.
*
*
*++ This function initializes the hardware of the CAN controller.
*++ It sets some global parameters like bitrate, interrupt settings
*++ and Message Masks.
*
*-- Diese Funktion initialisiert die Hardware des CAN Controllers.
*-- Sie setzt einige globale Parameter, wie Bitrate, Interrupt settings
*-- und Message Masks.
*
* .LP
*++ With CONFIG_FULLCAN not set,
*++ the object 15 is preconfigured to receive all CAN messages
*++ like a Basic CAN controller
*++ and the object 1 is preconfigured as a transmit object.
*
*-- Wenn CONFIG_FULLCAN nicht gesetzt ist, wird das Objekt 15
*-- vordefiniert zum Empfangen aller CAN Messages,
*-- wie bei einem BASIC CAN Controller,
*-- und das Objekt 1 wird vordefiniert als Transmit Objekt.
*
*++ Possible function prototyp for Init_CAN():
*-- Mglicher Funktionsprototyp fr Init_CAN():
* .CS
UNSIGNED8 Init_CAN(
    UNSIGNED16 wBaudRate,	 baudrate (e.g. 50 = 50kBaud)
    void *BaseAddr,		 TouCAN base address	
    UNSIGNED16 level,		 Used interrupt level within 68376
    UNSIGNED16 vector,		 Interrupt vector base
    UNSIGNED16 iarb		 Interrupt Arbitration value
    )
* .CE
* Parameters wBaseAddr, wClk, wTprop and bOutputMode
*++ are not used in this implementation.
*-- werden in dieser Implementierung nicht genutzt.
*
* RETURNS
* .TP
* \&CO_OK
*++ success
*-- Erfolg
* .TP
* \&CO_E_INIT_HARD_RES_ACTIVE
*++ reset is aktive, init failed
*-- Fehler, Hardware Reset ist aktiv
* .TP
* \&CO_E_INIT_BAUD
*++ could'nt adjust baudrate, init failed
*-- Fehler, gewmschte Baudrate kann nicht eingestellt werden oder
*-- ist fehlerhaft.
*
*/

UNSIGNED8 Init_CAN(
    UNSIGNED16 wBaudRate,	/* baudrate (e.g. 50 = 50kBaud) */
    void *BaseAddr,		/* TouCAN base address		*/
    UNSIGNED16 level,		/* Used interrupt level within 68376 */
    UNSIGNED16 vector,		/* Interrupt vector base */
    UNSIGNED16 iarb		/* Interrupt Arbitration value */
    CO_COMMA_LINE_PARA_DECL    )
{
UNSIGNED8	bN;

# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can;
# endif

#ifdef  CONFIG_DRIVER_TEST
    PRINTF(".. ->%s\n", __TIME__);
# ifdef CONFIG_FULLCAN
    PRINTF(".. def CONFIG_FULLCAN\n");
# else
    PRINTF(".. notdef CONFIG_FULLCAN\n");
# endif
#endif 		/* CONFIG_DRIVER_TEST */

    RESET_BIT(0);
    RESET_BIT(1);

    /* save the interrupt level for interrupt disable/enable functions */ 
    canIntLevel = (level << 8) & 0x0700;

    /* Init buffers */
#ifdef CONFIG_DRIVER_TRACE_ENABLE
    pTr = (UNSIGNED8 *)malloc(CONFIG_MAXTRACEBUFFER);
    pStTr = pTr;
#endif

    eSending CO_LINE_PARA_ARRAY_INDEX =
    eReceiving CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;

    /* Initialize  pointer before using CAN_ macros */
    tou_can = (tou_can_t *)BaseAddr;
# ifdef CONFIG_MULT_LINES
    tou_can_addr CO_LINE_PARA_ARRAY_INDEX = tou_can;
# endif
    
    /* reset TX/RX buffer */
    clearRxBuffer(CO_LINE_PARA);
    clearTxBuffer(CO_LINE_PARA);
    
    /*
     * go to INIT mode
     * Any configuration change/initialization requires that the TouCAN be
     * frozen by either asserting the HALT bit in the
     * module configuration register or by reset.
     * For Init_CAN() we choose reset.
     */
    CAN_RESET_BIT(CAN_ModulConfigRegister, CAN_MCR_FRZ);
    CAN_SET_BIT(CAN_ModulConfigRegister, CAN_MCR_SOFT_RST);
    SLOW_DOWN_IO();
    /* Test Reset Status */
    if(CAN_TEST_BIT(CAN_ModulConfigRegister, CAN_MCR_SOFT_RST) != 0) {
	return(CO_E_INIT_HARD_RES_ACTIVE);
    }

    {
    int mcr;
	mcr = CAN_READ(CAN_ModulConfigRegister);
	mcr &= 0xfff0;
	CAN_WRITE(CAN_ModulConfigRegister, (mcr + iarb));
    }
    CAN_WRITE(CAN_InterruptRegister, ((level << 8) + vector));

    /* Initialize the transmit and receive pin modes in control register 0 */
    CAN_WRITE(CAN_ControlReg0, 
    	  CAN_CTRL0_DISABLE_BOFF_INT	/* enable Bus-Off Interrupt	*/
    	+ CAN_CTRL0_DISABLE_ERR_INT	/* enable Error Interrupt	*/
    	+ 0				/*  4 , logic RX level		*/
    	+ 0				/*  1 , logic TX level		*/
    	);

    Set_Baudrate(wBaudRate, NULL CO_COMMA_LINE_PARA);	

    /*
     * Select the internal arbitration mode
     * LBUF in CANCTRL1
     * LBUF Lowest Buffer Transmitted First
     * The LBUF bit defines the transmit-first scheme.
     * 0 = Message buffer with lowest ID is transmitted first.
     * 1 = Lowest numbered buffer ist transmitted first.
     *
     * !! change the order of the next two statements  !!
     */
    CAN_SET_BIT(CAN_ControlReg1, CAN_CTRL1_LBUF);
    CAN_RESET_BIT(CAN_ControlReg1, CAN_CTRL1_LBUF);



    /*
     * Initialize message buffers.
     * The control/status word of all message buffers are written
     * as an inactive receive message buffer.
     */
    for(bN = 0; bN < 16; bN++) {
	CAN_WRITE_CTRL(bN, REC_CODE_NOT_ACTIVE, 0);
    }


    /*
     * Initialize mask registers for acceptance masks
     *
     * The receive mask registers are used as acceptance masks for received
     * frame IDs.
     * The following masks are defined:
     * - a global mask, used for receive buffers 0-13
     * - two separate masks for buffers 14 and 15
     * The value of the mask registers should not be changed during normal
     * operation. If the mask register data is changed
     * after the masked identifier of a received message is matched
     * to a locked message buffer, that message will be transferred
     * into that message buffer once it is unlocked,
     * regardless of wether that message's masked identifier still matches
     * the receive buffer identifier.
     * 
     * Mask bit values:
     *  0  - The corresponding incoming ID bit is "don't care".
     *  1  - The corresponding ID bit is checked against the incoming ID
     *       bit to see if a match exists.
     *
     * All mask are set for a full masking that means the received ID
     * must fully mach the message buffer mask
     * (as it is after reset).
     */

    CAN_WRITE(CAN_ReceiveGlobalMaskHigh,   0xFFEF);
    CAN_WRITE(CAN_ReceiveGlobalMaskLow,    0xFFFE);

    CAN_WRITE(CAN_ReceiveBuffer14MaskHigh, 0xFFEF);
    CAN_WRITE(CAN_ReceiveBuffer14MaskLow,  0xFFFE);

    CAN_WRITE(CAN_ReceiveBuffer15MaskHigh, 0xFFEF);
    CAN_WRITE(CAN_ReceiveBuffer15MaskLow,  0xFFFE);

    
#ifndef CONFIG_FULLCAN
    /*
     * The first object is preconfigured as a transmit object
     * with transmit interrupt enabled
     */
    CAN_SET_BIT(CAN_InterruptMasks,  (1 << TRANSMIT_OBJ));

    /*
     * Message object 14 should handle all CAN like Basic CAN.
     * That means all messages must arrive without masking,
     * all mask bits are set to don't care.
     */
    CAN_WRITE(CAN_ReceiveBuffer14MaskHigh, 0x0000);
    CAN_WRITE(CAN_ReceiveBuffer14MaskLow,  0x0000);

    /* Receive object also has interrupt enabled */
    CAN_SET_BIT(CAN_InterruptMasks,  (1 << ALL_RECEIVE_OBJ));
    CAN_WRITE_CTRL(ALL_RECEIVE_OBJ, REC_CODE_EMPTY, 8);
    CAN_WRITE_OID(ALL_RECEIVE_OBJ, 0);	/* set IDE - extended bit to zero */
#endif

    /*
     * Reset Pending Interrupts
     */
    CAN_READ(CAN_InterruptFlags);
    CAN_WRITE(CAN_InterruptFlags, 0);

#ifdef CONFIG_CAN_ERROR_HANDLING
    CAN_SET_BIT(CAN_ControlReg0,  CAN_CTRL0_ENABLE_BOFF_INT);
    CAN_SET_BIT(CAN_ControlReg0,  CAN_CTRL0_ENABLE_ERR_INT);
#endif

#ifdef CONFIG_CAN_TX_TEST
    /* configure and transmit one test object to measure the bitrate.
     * This could be done if no receiver is connected to the CAN network.
     * We use Msg-Id 100 with one data byte = 0xaa
     */
    CAN_WRITE_CTRL(TRANSMIT_OBJ, TRANS_CODE_NOT_READY, 1);

    CAN_WRITE_OID(TRANSMIT_OBJ, 100);

    CAN_OBJ[TRANSMIT_OBJ].MSG0 = 0xaa;

    CAN_WRITE_CTRL(TRANSMIT_OBJ, TRANS_CODE_TRANSMIT_ONCE, 1);

#endif
    /*
     * The HALT bit is set to one after reset.
     * It should be cleared after initializing the message buffers
     * and control registers.
     * TouCAN message buffer receive and transmit functions are inactive
     * until this bit is cleared.
     */
    CAN_RESET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT);

    return(CO_OK);
}

/*******************************************************************/
/**
*
*++ \brief Set_Baudrate - setting the CAN controlers bit rate 
*-- \brief Set_Baudrate - setzt den CAN-Controller auf die gewnschte Bitrate
*
*++ The function recalibrates the CAN controller with the given bit rate
*++ in K bit/s.
*++ The bit timing values for this bit rate are obtained from
*++ the table given as argument.
*++ If no table is given, the default table is used.
*++ After finishing the function, the CAN controller is left stopped.
*
*-- Diese Funktion stellt den CAN-Controller auf die bergebene
*-- Bitrate (in kBit/s) ein.
*-- Die dazugehrigen Bittimings werden
*-- der bergebenen Tabelle entnommen.
*-- Wird keine solche Tabelle bergeben,
*-- wird die vordefinierte Tabelle benutzt. Nach Ausfhrung
*-- der Funktion befindet sich der CAN-Controller im gestoppten
*-- Zustand.
*
* \retval CO_INIT_CAN_OK
*++ success
*-- Erfolg
* \retval CO_E_INIT_BAUD
*++ could'nt adjust baudrate, init failed
*-- Fehler, ungltige Baudrate
*
* \internal
* eigentlich sollte die Tabelle doch entsprechend cia sortiert sein,
* d.h. es gibt einen Tabellen Index von 0...8 (oder 7??)
* und nicht eine Bitrate in kbit/s
*/

UNSIGNED8 
Set_Baudrate(
	UNSIGNED16 rate, 	/* Baudrate (125 == 125kBit/s )*/
	BTR_TAB_T *tab		/* NULL == interne Tabelle */
CO_COMMA_LINE_PARA_DECL) 
{
# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif
	
BTR_TAB_TOUCAN_T * table = (BTR_TAB_TOUCAN_T*)tab;


    /* enable changing of bit timings */
    CAN_SET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT);

    if (rate == 0) {    	/* Baudrate 0 isn't allowed */
    	return( CO_E_INIT_BAUD );
    }

    if (table == NULL) {  	/* default table for Bittimings */
    	table = can_btr_tab_toucan;  
    }	
    /* search data from table */
    while(1) {
        if (table->rate == 0) {
            return( CO_E_INIT_BAUD );
        }    
    	if (table->rate == rate) {
    	    break;
    	}	
    	table++;
    }
    
    /*
     * Set Timing Register values.
     * Initialize the bit timing parameters PROPSEG, PSEG1, PSEG2 and RJW
     * in control registers 1 and 2.
     *
     * The TouCAN module uses three 8-bit registers to set-up
     * the bit timing parameters required by the CAN protocol.
     * Control registers 1 and 2 contain the PROPSEG, PSEG1, PSEG2
     * and RJW fields which allow the user to configure
     * the bit timing parameters.
     * The prescaler divide register PRESDIV) allows the user to select
     * the ratio used to derive the S-Clock from the system clock.
     * The time quanta clock operates at the S-clock frequency. 
     */
    CAN_WRITE(CAN_PrescalerDividerRegister, table->presdiv);
    CAN_WRITE(CAN_ControlReg2, ((table->pseg1 << 3)+ table->pseg2));
    CAN_WRITE(CAN_ControlReg1, (
    	  0		/* SAMP = 0 , 0/0x80		*/	
    	+ 0		/* LOOP = 0 , 0/0x40		*/	
    	+ 0		/* TSYNC= 0 , 0/0x20		*/
    	+ 0		/* LBUF = 0 , 0/0x10		*/
    	+ 0		/* RSVD = 0 , 0/0x08		*/
    	+ table->propseg)
    	);

    /* Im gestoppten Zustand bleiben -> Start_CAN() aktiviert den Controller
     * mit der neuen Baudrate
     */
    return(CO_INIT_CAN_OK);
}


/*******************************************************************
*
* Start_CAN - starts the CAN-Controller
*
* Starts the CAN-Controller by resetting the HALT bit.
*
* RETURNS
* .TP
* nothing
*
*/

void Start_CAN(CO_LINE_PARA_DECL
	)
{
# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif
	
    CAN_Interrupts_allowed CO_LINE_PARA_ARRAY_INDEX = CO_TRUE;
    eSending CO_LINE_PARA_ARRAY_INDEX =
    eReceiving CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
    CAN_RESET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT);
    ENABLE_CAN_INTERRUPTS(CO_LINE_PARA);

}

/*******************************************************************
*
*++ Stop_CAN - stops the CAN-Controller
*-- Stop_CAN - Stoppt den CAN-Controller 
*
*++ Stops the CAN-Controller by setting the HALT bit.
*-- Stoppt den CAN Controller durch setzen des HALT bits.
*
* RETURNS
* .TP
*++ nothing
*-- nichts
*
*/

void Stop_CAN(CO_LINE_PARA_DECL
	)
{
# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif

    CAN_SET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT);
}

/*******************************************************************/
/**
*
*++ \brief Clear_busoff - starts the CAN-Controller after bus-off
*-- \brief Clear_busoff - startet den CAN Controller nach einem Bus OFF
*
*++ In case of a Bus-OFF the controller is in init-reset mode.
*++ We only must clear the init bit to go bus on again.
*-- Im Fall das der CAN Controller in den Bus-Off Zustand gegangen ist,
*-- Init-Reset Bit ist gesetzt,
*-- kann mit Aufruf dieser Funktion durch Rcksetzen des Init-Bit
*-- versucht werden, den CAN Controller wieder an den Bus zu bringen.
*
*++ If this call is used more times successive,
*++ a waiting time between the calls should be used.
*-- Bei wiederholten Aufrufen, sollte eine Wartezeit zwischen
*-- den Aufrufen liegen.
*
* \returns
*++ nothing
*-- nichts
*/

void Clear_busoff(CO_LINE_PARA_DECL
     )
{
# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif

    /* ?? *//* CAN_SET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT); */
    
    eSending CO_LINE_PARA_ARRAY_INDEX =
    eReceiving CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
    CAN_RESET_BIT(CAN_ModulConfigRegister, CAN_MCR_HALT);
		  
}

/*******************************************************************
*
*++ Define_COB - creates a COB in the COB-list with attributes
*-- Define_COB - erzeugt ein COB mit Attributen in der COB-Liste 
*
*++ Creates a COB in the COB-list with attributes given as parameter.
*++ If it is the first call the list is created otherwise
*++ the COB entry is appanded to the list.
*++ In both cases memory is allocated with a call to 
*++ .I CalMalloc(2) .
*-- Erzeugt ein COB mit den als Parameter gegebenen Attributen
*-- und fgt es in die COB-Liste ein. Wenn es der erste Aufruf
*-- ist, wird die Liste kreiert, andernfalls wird das COB der
*-- Liste hinzugefgt.
*-- In beiden Fllen wird Speicher durch einen Aufruf von
*-- .I CalMalloc(2)
*-- angefordert.
*
*++ With Full-CAN controllers also object channels in the controllers hardware
*++ are occupied.
*++ The Channel is configured according to the Type of the COB
*++ as transmit or receive object
*++ The COB ID would be assigned later, with a call to Set_COB_ID().
*-- Mit Full-CAN Controllern werden auch die Objekt-Channels in der
*-- Controllerhardware belegt.
*-- Der Channel wird entsprechend dem Typ des COB als Transmit
*-- oder Receive Objekt konfiguriert.
*-- Die COB ID wird spter mit einem Aufruf von Set_COB_ID() eingestellt.
*
*++ If CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
*++ is set, only one channel is used for all transmit objects
*++ plus one additional transmit channel for Node-Guarding.
*++ If this is enabled, transmit channels can !!! not !!! handle RTR Requests.
*-- Wenn CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
*-- gesetzt ist, wird nur ein Channel fr alle Transmit Objekte benutzt.
*-- Zustzlich wird noch ein Transmit Channel fr das Node-Guarding
*-- benutzt. Wenn dieses Define gesetzt ist, knnen Transmit-Channels
*-- !! keine !! RTR Requests bearbeiten.
*
* RETURNS
* .TP
*++ pointer to COB
*-- Zeiger auf das COB
* \&OK
* .TP
* 0
*++ Error, no more channels within Full-CAN controller
*-- Fehler, keine freien Channels mehr im Full-CAN Controller
*
*/
/* EZ, was: struct CO_COB *Define_COB(  */
COB_T *Define_COB( 
	COB_KIND_T eType,  /* COB type */
	UNSIGNED8 bLength         /* COB length*/
	CO_COMMA_LINE_PARA_DECL)
{
#  if defined(CONFIG_FULL_CAN) && defined(CONFIG_MULT_LINES)
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif

struct CO_COB 	*pCOB_InUse,
		*pOld_COB_InUse;

#ifdef CONFIG_FULLCAN
/* organize channel using of CAN controller */
/* Multiline support for only one transmit channel still missing */ 
# ifdef CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
/*
 * First channel is reserved for all transmit objects
 * Second Channel, if configured with CONFIG_NODE_GUARDING
 * for that.
 */
#  if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
static UNSIGNED8 bChannel = 2;	/* starting with the first free one */
#  else				/* after transmit and node guard channel */
static UNSIGNED8 bChannel = 1;	/* starting with the first free one */
#  endif				/* after transmit channel           */
# else /* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL */
# ifdef CONFIG_MULT_LINES
/* the initializer list has to be changed according to the value of CONFIG_MULT_LINES: */
static UNSIGNED8 bChannel_arr CO_LINE_PARA_ARRAY_DEF = {0,0}; 
UNSIGNED8 bChannel = bChannel_arr CO_LINE_PARA_ARRAY_INDEX; 
#else
static UNSIGNED8 bChannel = 0;	/* starting with the first free one */
# endif         /* CONFIG_MULT_LINES */
# endif 	/* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL */

# ifdef CONFIG_DRIVER_TEST
   PRINTF("\nDefine_COB(): ");
# endif
//    printf2("Define_COB, for line ",canLine);
    if (bChannel  == 14) {
        /* No more channels available */
#       ifdef CONFIG_DRIVER_TEST
        PRINTF("No more Message Objects\n");
#       endif
    	return ((struct CO_COB *)(0));
    }
#endif 		/* CONFIG_FULLCAN */

    if ((pCOB_InUse = pFirst_COB_Entry CO_LINE_PARA_ARRAY_INDEX) == NULL)
    {
	if(( pCOB_InUse =
	     pFirst_COB_Entry CO_LINE_PARA_ARRAY_INDEX =
		CalMalloc(sizeof(struct CO_COB))) == NULL) {
	    return(NULL);
	}
	pFirst_COB_Entry CO_LINE_PARA_ARRAY_INDEX ->pNext = NULL;
    }
    else
    {
	while (pCOB_InUse->pNext != NULL)
	{
	    pCOB_InUse = pCOB_InUse->pNext;
	}
	pOld_COB_InUse = pCOB_InUse;
	if(( pOld_COB_InUse->pNext =
	     pCOB_InUse = CalMalloc(sizeof(struct CO_COB))) == NULL)
	{
	    return(NULL);
	}
	pCOB_InUse->pNext = NULL;
    }

#if defined(CONFIG_DRIVER_TEST)
/*
RX_COB		receive message
RTR_RX_COB	receive message, which can be requested by the local device
TX_COB		transmit message
RTR_TX_COB	transmit message, which can be requested by remote devices

That means that both TX_COB and RTR_TX_COB are data messages to transmit.
That other can be of type RTR_RX_COB,
that means transmit a RTR message from local node.
*/
    switch(eType) {
	case RTR_TX_COB:
		PRINTF("RTR_TX_COB\n"); break;
	case RX_COB    :
		PRINTF("RX_COB\n"); break;
	case TX_COB    :
		PRINTF("TX_COB\n"); break;
	case RTR_RX_COB:
		PRINTF("RTR_RX_COB\n"); break;
	default:
		PRINTF("!!! unknown COB type !!!\n"); break;
    }
#endif

    pCOB_InUse->bLength  = bLength;
    pCOB_InUse->eType    = eType;

# ifdef CONFIG_MULT_LINES
    pCOB_InUse->canLine  = canLine;
# endif


#ifdef CONFIG_FULLCAN

    /* The Channel must be configured according to the Type of the COB
     * The COB ID would be assigned later, with a call to Set_COB_ID().
     */
    if (eType == RX_COB || eType == RTR_RX_COB) {
	CAN_WRITE_CTRL(bChannel, REC_CODE_NOT_ACTIVE, 0);

	pCOB_InUse->bChannel	= bChannel;
    }
    else {	/* it is a transmit type */

# ifdef CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
	/*
	 * all transmit objects are assigned to one channel,
	 * the TRANSMIT_OBJ channel. It is already configured in
	 * init_CAN().
	 * If CONFIG_NODE_GUARDING is set,
	 * later on in set_COB_ID()
	 * the node guarding channel is reassigned to the
	 * NODEGUARD_OBJ channel.
	 */
	pCOB_InUse->bChannel	= TRANSMIT_OBJ;

# else	/* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL*/

	pCOB_InUse->bChannel	= bChannel;
	CAN_WRITE_CTRL(bChannel, TRANS_CODE_NOT_READY, bLength);

# endif		/* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL*/
    }

    /* set interrupt enable for this channel */
    CAN_SET_BIT(CAN_InterruptMasks, (1 << pCOB_InUse->bChannel));

# ifndef CONFIG_ONLY_ONE_TRANSMIT_CHANNEL
    /*
     *increment message object/channel counter at the end,
     * because we want to start with index 0 in CAN_OBJ[]
     */
    bChannel++;
# else   /* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL*/ 
    if (eType == RX_COB || eType == RTR_RX_COB) {
        bChannel++;
    }    
# endif	/* CONFIG_ONLY_ONE_TRANSMIT_CHANNEL*/

#endif 		/* CONFIG_FULLCAN */

#  if defined(CONFIG_FULL_CAN) && defined(CONFIG_MULT_LINES)
    bChannel_arr CO_LINE_PARA_ARRAY_INDEX = bChannel;
# endif
//    printf2("Define_COB, result: ",pCOB_InUse);
    return(pCOB_InUse);
}


/*******************************************************************
*
*++ Set_COB_ID - assigns an identifier to a specific COB
*-- Set_COB_ID - bertrgt eine ID zu einem spezifischen COB 
*
*++ Uses the Channel information to change the identifier of this channel.
*-- Benutzt die Channel Information zum ndern der ID fr diesen Channel.
*
* RETURNS
* .TP
*++ nothing
*-- nichts
*
*/

void Set_COB_ID(
     struct CO_COB * pCOB, /* pointer to COB in list */
     UNSIGNED16 wID           /* identifier */
     )
{
#ifdef CONFIG_FULLCAN
UNSIGNED8 bChannel;		/* COB-IDs channel number */
#endif

    if (pCOB != NULL) {
#if defined(CONFIG_FULL_CAN) && defined(CONFIG_MULT_LINES)
    UNSIGNED8 canLine = pCOB->canLine;
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
#endif
//    printf2("Set_COB_ID, pCOB= ",pCOB);
//    printf2("            wID= ",wID);    	
	pCOB->wID = wID;

#ifdef CONFIG_FULLCAN

	bChannel = pCOB->bChannel;

# ifdef CONFIG_DRIVER_TEST
	PRINTF("Set_COB_ID(), Channel=%d, Type=%d, id=%d/%x\n",
	    (UNSIGNED16)bChannel,
	    (UNSIGNED16)pCOB->eType,
	    (UNSIGNED16)wID,
	    (UNSIGNED16)wID);
# endif 	/* CONFIG_DRIVER_TEST */
//   printf2("Set_COB_ID, etype= ",pCOB->eType);
//   printf2("            wID= ",wID);

# if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
	if(    pNode CO_LINE_PARA_ARRAY_INDEX !=  NULL
	    && pNode CO_LINE_PARA_ARRAY_INDEX->pGuard_COB != NULL 
	    && wID == pNode CO_LINE_PARA_ARRAY_INDEX->pGuard_COB->wID) {
#  ifdef CONFIG_DRIVER_TEST
	    PRINTF("  Yeeeaahp, it's Node Guarding Channel\n");
#  endif
#  if defined(CONFIG_ONLY_ONE_TRANSMIT_CHANNEL)
	    pCOB->bChannel = bChannel = bNodeGuardingChannel = NODEGUARD_OBJ;
#   ifdef CONFIG_DRIVER_TEST
	    PRINTF("  give ID to Ch %d\n", (UNSIGNED16)NODEGUARD_OBJ);
#   endif
#  else
	    bNodeGuardingChannel  CO_LINE_PARA_ARRAY_INDEX	= bChannel;
#  endif
	}
# endif
# if (defined(CONFIG_SYNC_CONSUMER) || defined(CONFIG_SYNC_PRODUCER))
	if(    pSync !=  NULL
	    && pSync->pCobId != NULL
	    && wID == pSync->pCobId->wID) {
#  ifdef CONFIG_DRIVER_TEST
	    PRINTF("  Yeeeaahp, it's the SYNC Channel\n");
#  endif
	    bSyncChannel		= bChannel;
	}
# endif

//   printf2("Set_COB_ID, using channel ",bChannel);
    /* initialization for transmit and receive buffers are differ */
    if ((pCOB->eType == RX_COB) || (pCOB->eType == RTR_RX_COB)) {
	/* Reset Message Valid */
	CAN_WRITE_CTRL(bChannel, REC_CODE_NOT_ACTIVE, 0);

	if (pCOB->eType == RTR_RX_COB) {
	    CAN_WRITE_OID_RTR(bChannel, wID);
	} else  {
	    CAN_WRITE_OID(bChannel, wID);
	}

	/* set buffer ready */
	CAN_WRITE_CTRL(bChannel, REC_CODE_EMPTY, pCOB->bLength);
    } else { /* is a transmit object */
	/* Reset to not ready */
	CAN_WRITE_CTRL(bChannel, TRANS_CODE_NOT_READY, 0);
	CAN_WRITE_OID(bChannel, wID);
    }

#endif		/* CONFIG_FULLCAN */

    }	/* (pCOB != NULL) */
}


#ifdef CONFIG_FULLCAN
/*******************************************************************
*
*++ Update_COB - updates a transmit COBs data
*-- Update_COB - aktualisiert die Daten eines Transmit COBs
*
*++ Updates the data contents of a COB in a full CAN controller
*++ with the data
*++ .I pMsg
*++ points to.
*-- Aktualisiert den Inhalt der Daten in einem Full CAN Controller
*-- mit den Daten auf die
*-- .I pMsg
*-- zeigt.
*
*++ This function is only useful on a Full CAN-Controller
*++ on a Basic CAN it is a dummy function.
*++ We use this function only for nodeguarding function !!
*-- Diese Funktion ist nur bei Full CAN Controllern sinnvoll,
*-- bei Basic CAN ist dies nur eine Dummy Funktion.
*-- Wir benutzen diese Funktion nur fr die Node Guarding Funktion!!
*
* RETURNS
* .TP
*++ nothing
*-- nichts
*
*/

void Update_COB(
     COB_T * pCOB,		/* pointer to COB in list*/
     UNSIGNED8 * pMsg		/* data */
     )
{
UNSIGNED8 bChannel;		/* actual Channel number from 0...14 */
UNSIGNED16 channel_state;	/* actual status of channel */
# if !defined( CONFIG_DRIVER_USE_MEMCPY )
UNSIGNED8 i;
# endif
    bChannel = pCOB->bChannel;

# ifdef CONFIG_DRIVER_TEST
	 putchar('.'); fflush(stdout); 
# endif
    if (pCOB != NULL)
    {
#ifdef CONFIG_MULT_LINES
    UNSIGNED8 canLine = pCOB->canLine;
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
#endif

/*    	printf1("Update_COB running"); */
/*      EZ Test; to be confirmed or removed: */
/*    	DISABLE_CAN_INTERRUPTS(CO_LINE_PARA); */
# if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
	if (bChannel == bNodeGuardingChannel  CO_LINE_PARA_ARRAY_INDEX) {
	    /* Reset to not ready */
	    CAN_WRITE_CTRL(bChannel, TRANS_CODE_NOT_READY, 0);

	/* Update data */
#  if defined(CONFIG_DRIVER_USE_MEMCPY )
	    CO_MEMCPY((UNSIGNED8*)&CAN_OBJ[bChannel].msg, pMsg, pCOB->bLength);
#  else
	    for(i = 0; i < pCOB->bLength; i++) {
	        *((UNSIGNED8 *)(&CAN_OBJ[bChannel].msg) + 
	         (sizeof(struct can_obj) * i))
	    		= *(pMsg++);	
	    }
#  endif

	    /* send only after received a RTR frame */
	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONLY_RTR,pCOB->bLength);
	}
/* TEST */
        else if (pCOB->eType == RTR_TX_COB){
# endif /* CONFIG_NODE_GUARDING */
            do{
                channel_state = CAN_READ_CTRL(bChannel);
                channel_state = channel_state & 0x000F;
            }while(channel_state == TRANS_CODE_TRANSMIT_ONCE_RTR_ALWAYS);
	    CAN_WRITE_CTRL(bChannel, TRANS_CODE_NOT_READY, 0);
	    CO_MEMCPY((UNSIGNED8*)&CAN_OBJ[bChannel].msg, pMsg, pCOB->bLength);
	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONLY_RTR,pCOB->bLength);
# if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
        }
# endif /* CONFIG_NODE_GUARDING */
/* **** */
/* did not help:    ENABLE_CAN_INTERRUPTS(CO_LINE_PARA); */
    }
#  ifdef CONFIG_DRIVER_TEST
    else {
	PRINTF("  Error: pCOP == NULL\n");

    }
#  endif
}
#endif		/* CONFIG_FULLCAN */


/*************************************************************************
*
* CAN_int - interrupt service for CAN controller
*
*
* The ISR subroutine contains actually many differnet implementation
* variants.
* All are selectable with compiler constants.
* .LP
* First there is a general decision to use the
* TouCAN as a Basic CAN or Full CAN controller.
* The Full CAN implemntation supports all types of CAN frames
* but is limited in CANopen to handle only up to 16 message objects.
* This is always enough for simple devices.
* The following can be an example.
* .LP
* .nf
* - Network NMT Message (must)
* - one SDO Message pair (must)
* - one PDO pair
* - one emergency message
* - one node guarding Message
* -----------------------------
* 7 message objects
*
* .LP
* If SYNC Services are defined in
* .I cal_conf.h
* the following further compiler constants are possible
* .CS
* CONFIG_UPDATE_SYNC_ACTUAL_MSG
* CONFIG_UPDATE_SYNC_CMD_MSG
* .CE
* .LP
* If CONFIG_UPDATE_SYNC_CMD_MSG
* is set,
* a user function
* .I updateSyncCmdMsg(UNSIGNED16 pdoNr)
* is called from within the CAN ISR.
* With this function the user can update his variables mapped into
* a sync RPDO after receiving it with SYNC.
* .LP
* If CONFIG_UPDATE_SYNC_ACTUAL_MSG
* is set,
* a user function
* .I updateSyncActualMsg(UNSIGNED16 pdoNr, UNSIGNED8 *pData)
* is called from within the CAN ISR.
* With this function the user can update his variables mapped into
* a sync TPDO before sending it with SYNC.
* .LP
* .I updateSyncCmdMsg()
* is called before
* .I updateSyncActualMsg() .
* 
* In 8 bit applications the ISR consumes more time then
* the repetition rate of more as two consecutive interrupts with no data
* and the maximum bit rate of 1 Mbit/s.
* Tis is one of the reasons to put such things like bit shifting for
* the message ID, outside of the ISR.
* If the CONFIG_C515C is set, this is done in FlushMbox(), not in the ISR.
* Reading and writing of the message Id object register is therefore
* put in a Macro, called CAN_READ_OID(obj).
*
* INTERNAL
*
* For time measurement PIN 1 is used to measure the length of the ISR.
* The Pin is set to High when entering the ISR after all ISR entering code
* provided by the compiler and Reset to Low bevor executing the restoring code.
* Within the ISR PIN1
* is forced to Low while transfering data from the CAN controllers
* data registers to the CAL receive puffer.
*
*
*        Int     Entering C-ISR                           Leaving ISR
*        |         |                                       |
*        v         v                                       v
*                 ____________            ___________
* _______________|           |___________|          |_______________
*                           
*                           ^           ^
*                            data transfer
*
*
*
* The following subroutines are used within ISR and partially from main.
* updateSyncRpdo()			using
* 		updateSyncCmdMsg()	using
* 		insertMbox()		NOAREG
* transSyncPdo()				using
* 		updateSyncActualMsg()	using
* 		Insert_TX_Request()	NOAREG
* 		GetNext_TX_Request()	NOAREG
* 
* GetNext_TX_Request()			NOAREG
* insertMbox()				NOAREG
* 
*/
/* For the use in the Ada Context this routine is called
   from the protected procedure which is the actual ISR
   for the CAN interrupt */  /* EZ */
void CAN_int (
	CO_LINE_PARA_DECL) /* CAN_ISR_NUMBER CAN_ISR_REGISTERBANK*/
{
# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif

#ifdef CONFIG_FULLCAN
UNSIGNED16	wStat;	/* Interruptstate */
UNSIGNED16	wCh = 0;/* channel number */
#else
//UNSIGNED8 bLength;
//UNSIGNED8 bStat;	/* holds the contents of the status register */
//UNSIGNED8 bChannel;	/* holds the contents of the interrupt register */
int last_tx_id;
#endif

// PUSH_ALL_REGS(); 

//    SET_BIT(1);			/* start measuring ISR time */
    DISABLE_CAN_INTERRUPTS(CO_LINE_PARA);	/* disable CAN interrupt */
//    printf1("CAN_int, interrupts disabled");

#ifdef CONFIG_FULLCAN
    /* In this version, we use only one interruptroutine for all int-sources */
    /* therefore we have to select the int-source here */
    if (CAN_TEST_BIT(CAN_ErrorStatusRegister,
	(CAN_ESTAT_BOFF_INT + CAN_ESTAT_ERR_INT + CAN_ESTAT_WAKE_INT)))  {

        eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;

# ifdef  CONFIG_DRIVER_TEST
		    cal_ch_ints++;
# endif 	    /* CONFIG_DRIVER_TEST */

# ifdef CONFIG_CAN_ERROR_HANDLING
	/* there is a interruptbit set */
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_BOFF_INT))  {
	    /* Bus off int */
	    SET_COLIB_FLAG(COFLAG_CAN_BUSOFF);
	}
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_ERR_INT))  {
	    /* error int */
	    SET_COLIB_FLAG(COFLAG_CAN_PASSIVE);
	}
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_WAKE_INT))  {
	    /* wake int */
	    SET_COLIB_FLAG(COFLAG_CAN_PASSIVE);
	}
# endif /* CONFIG_CAN_ERROR_HANDLING */(

	/* reset int source */
	CAN_RESET_BIT(CAN_ErrorStatusRegister,
	    (CAN_ESTAT_BOFF_INT + CAN_ESTAT_ERR_INT + CAN_ESTAT_WAKE_INT));
    }

    /* it is possible we have a new interrupt while we work here */
    while ((wStat = CAN_READ(CAN_InterruptFlags)))  {

	/* reset the int bits */
	CAN_WRITE(CAN_InterruptFlags, 0);

	/* shift the interruptregister while > 0 */
	while (wStat != 0)  {
	    /* if this bit set ?? */
	    if (wStat & 1)  {
		/* read state from this channel */
		if (CAN_READ_CTRL(wCh) & TRANS_CODE_NOT_READY)  {
		    /* we have a transmit interrupt */
# ifdef  CONFIG_DRIVER_TEST
		    cal_tx_ints++;
# endif 	    /* CONFIG_DRIVER_TEST */
//                    printf2("CAN_int, got transmit interrupt  , ctrl=  ",CAN_READ_CTRL(wCh)& 0x0f);


# if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
		    if (wCh == bNodeGuardingChannel CO_LINE_PARA_ARRAY_INDEX) {
			/* CAN has answered of the Nodeguarding RTR Message */
			
			if (pRX_Buffer[bRX_WriteIndex  CO_LINE_PARA_ARRAY_INDEX] 
			CO_LINE_PARA_ARRAY_INDEX .eStat != FULL) {
			    /* receive queue is free */
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
			    CO_LINE_PARA_ARRAY_INDEX .wCOB_ID =
			    	CAN_READ_OID(wCh);
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
			    CO_LINE_PARA_ARRAY_INDEX .bLength =
			        CO_RTR_REQ;
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
			     CO_LINE_PARA_ARRAY_INDEX .pData[0] =
			        (UNSIGNED8)CAN_OBJ[wCh].MSG0;
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
			     CO_LINE_PARA_ARRAY_INDEX .eStat = FULL;
			    if (++bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX == CONFIG_RX_BUFFER_SIZE) {
				bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX = 0;
			    }
			}
			eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
		    } else {
# endif /* CONFIG_NODE_GUARDING */

			eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
			GetNext_TX_Request(CO_LINE_PARA); /* get next transmission request */
# if defined(CONFIG_NODE_GUARDING) && defined(CONFIG_SLAVE)
		    }
# endif /* CONFIG_NODE_GUARDING */

		} else {

		    /* we have a receive interrupt */
# ifdef  CONFIG_DRIVER_TEST
		    cal_rx_ints++;
# endif 	    /* CONFIG_DRIVER_TEST */

	    	/* ez */
//		printf2("CAN_int, got receive interrupt, ctrl=  ",CAN_READ_CTRL(wCh)& 0x0f);
                while ((CAN_READ_CTRL(wCh) & REC_CODE_BUSY) == REC_CODE_BUSY) {
//                	printf1("CAN_int, REC_CODE_BUSY");
		};
                /* 20 cycles maximum wait */
                if ((CAN_READ_CTRL(wCh) & REC_CODE_FULL) != REC_CODE_FULL)
                {
//                   printf2("CAN_int, buffer not FULL, ctrl= ",CAN_READ_CTRL(wCh));
                } else {
		    /*
		     * First handle all Messages which must not be sent to
		     * the upper layer, but are handled within the ISR:
		     * SYNC - is a RX_COB
		     *
		     */
# if (defined(CONFIG_SYNC_CONSUMER) || defined(CONFIG_SYNC_PRODUCER))
		    if (wCh == bSyncChannel CO_LINE_PARA_ARRAY_INDEX ) {
			/* SYNC message received */
			/* set values from shadow buffer valid */
			SET_COLIB_FLAG(COFLAG_SYNC_RECEIVED);
		    }
		    else {
# endif /* defined(CONFIG_SYNC_CONSUMER) || defined(CONFIG_SYNC_PRODUCER) */
                   if ((wCh == 9) & (eSending CO_LINE_PARA_ARRAY_INDEX == CO_TRUE)) /* EZ */ {
                   	eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
//       			printf1("CAN_int, reset eSending");

                   }	
			/*
			 * Second: all Messages which has to be put
			 * in the receive queue.
			 */
			if (pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
				CO_LINE_PARA_ARRAY_INDEX .eStat == FULL) {
			    /* no more space in receive queue */
# ifdef CONFIG_CAN_ERROR_HANDLING
			    SET_COLIB_FLAG(COFLAG_CAN_OVERFLOW);
# endif /* CONFIG_CAN_ERROR_HANDLING */
			} else {
			    /* receive queue is free */
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
				CO_LINE_PARA_ARRAY_INDEX .wCOB_ID =
			    	CAN_READ_OID(wCh); 
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
				CO_LINE_PARA_ARRAY_INDEX .bLength =
				CAN_OBJ[wCh].ctl_status & 0xf;

# if defined(CONFIG_DRIVER_USE_MEMCPY )
			    CO_MEMCPY((void*)&pRX_Buffer[bRX_WriteIndex
				CO_LINE_PARA_ARRAY_INDEX]
				CO_LINE_PARA_ARRAY_INDEX .pData[0],
				(UNSIGNED8 *)&CAN_OBJ[wCh].MSG0, 8);
# else
			    for(i = 0; i < pCOB->bLength; i++) {
				*(pMsg++) =
				*((UNSIGNED8 *)(&CAN_OBJ[bChannel].msg) + 
				(sizeof(struct can_obj) * i));
			    }
# endif
			    pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
				CO_LINE_PARA_ARRAY_INDEX .eStat = FULL;
			    if (++bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX
				== CONFIG_RX_BUFFER_SIZE) {
				bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX = 0;
			    }
	    	        }
# if (defined(CONFIG_SYNC_CONSUMER) || defined(CONFIG_SYNC_PRODUCER))
		    }
# endif /* defined(CONFIG_SYNC_CONSUMER) || defined(CONFIG_SYNC_PRODUCER) */
                    
		    /* set buffer ready for RX buffers */
		    CAN_WRITE_CTRL(wCh, REC_CODE_EMPTY, CAN_OBJ[wCh].ctl_status & 0xf);
		    }
		}
	    }
	    wStat >>= 1;	/* shift right */
	    wCh++;		/* next channel */
	}
    }

#else
/* --------------- ! CONFIG_FULLCAN -- BASIC CAN ----------------------------*/
/*===========================================================================*/

    if (CAN_TEST_BIT(CAN_ErrorStatusRegister,
	(CAN_ESTAT_BOFF_INT + CAN_ESTAT_ERR_INT + CAN_ESTAT_WAKE_INT)))  {

# ifdef CONFIG_CAN_ERROR_HANDLING
	/* there is a interruptbit set */
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_BOFF_INT))  {
	    /* Bus off int */
	    SET_COLIB_FLAG(COFLAG_CAN_BUSOFF);
	}
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_ERR_INT))  {
	    /* error int */
	    SET_COLIB_FLAG(COFLAG_CAN_PASSIVE);
	}
	if (CAN_TEST_BIT(CAN_ErrorStatusRegister, CAN_ESTAT_WAKE_INT))  {
	    /* wake int */
	    SET_COLIB_FLAG(COFLAG_CAN_PASSIVE);
	}
# endif /* CONFIG_CAN_ERROR_HANDLING */(

	/* reset int source */
	CAN_RESET_BIT(CAN_ErrorStatusRegister,
	    (CAN_ESTAT_BOFF_INT + CAN_ESTAT_ERR_INT + CAN_ESTAT_WAKE_INT));
    } 
/* Receive INT -------------------------------------------------------------*/
    if(CAN_TEST_BIT(CAN_InterruptFlags, (1 << ALL_RECEIVE_OBJ))) {
# ifdef  CONFIG_DRIVER_TEST
	cal_rx_ints++;
# endif 		/* CONFIG_DRIVER_TEST */
 
    	/* wdh */
        while ((CAN_READ_CTRL(TRANSMIT_OBJ) & REC_CODE_BUSY) == REC_CODE_BUSY) {
        	/* printf1("CAN_int, tx REC_CODE_BUSY"); */
	};
	
	last_tx_id = CAN_READ_OID_AND_RTR(TRANSMIT_OBJ);
	
	
    	/* wdh */
        while ((CAN_READ_CTRL(ALL_RECEIVE_OBJ) & REC_CODE_BUSY) == REC_CODE_BUSY) {
        	/* printf1("CAN_int, rx REC_CODE_BUSY"); */
	    };
        /* 20 cycles maximum wait */

	if (CAN_READ_OID_AND_RTR(ALL_RECEIVE_OBJ) == last_tx_id)
        {
          /* printf2("CAN_int, got fake rx int : OID= ", CAN_READ_OID(ALL_RECEIVE_OBJ)); */
          /*
	    This is obviously a message we sent ourselves recently.
	    There is no point in delivering this to the upper layer.
	    We rely on the fact that no other node will ever send a message with the same OID.
           */
        }else{
	   /* printf2("CAN_int, got true receive interrupt : OID= ",CAN_READ_OID(ALL_RECEIVE_OBJ)); */
	
	/* pCOB->wID = CAN_READ_OID(ALL_RECEIVE_OBJ); */
	/* bLength = CAN_OBJ[ALL_RECEIVE_OBJ].ctl_status & 0xf; */
	/* pCOB->bLength = bLength; */
	/* CO_MEMCPY(pMsg, (void *)&CAN_OBJ[ALL_RECEIVE_OBJ].MSG0, bLength); */
	    if (pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
	              CO_LINE_PARA_ARRAY_INDEX .eStat == FULL)
	    {
	    }
	    else
	    {
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .wCOB_ID =	CAN_READ_OID(ALL_RECEIVE_OBJ);
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .bLength =	CAN_OBJ[ALL_RECEIVE_OBJ].ctl_status & 0x0f;
		CO_MEMCPY(pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .pData, &CAN_OBJ[ALL_RECEIVE_OBJ].MSG0, 8);
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		      CO_LINE_PARA_ARRAY_INDEX .eStat = FULL;
//              printf2("received OID= ",pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
//		    CO_LINE_PARA_ARRAY_INDEX .wCOB_ID  );

		if (++bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX == CONFIG_RX_BUFFER_SIZE)
		{
		    bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX = 0;
		}
	    }
	}
	
	/*  reset pendin interrupt and activate receive buffer */
	CAN_RESET_BIT(CAN_InterruptFlags, (1 << ALL_RECEIVE_OBJ));
	CAN_WRITE_CTRL(ALL_RECEIVE_OBJ, REC_CODE_EMPTY, 8);
	CO_NEW_RX_MSG(CO_LINE_PARA);
    }

/* Transmit INT -------------------------------------------------------------*/
    if(CAN_TEST_BIT(CAN_InterruptFlags, (1 << TRANSMIT_OBJ))) {
# ifdef  CONFIG_DRIVER_TEST
	cal_tx_ints++;
# endif 		/* CONFIG_DRIVER_TEST */

    	/* wdh */
        while ((CAN_READ_CTRL(TRANSMIT_OBJ) & REC_CODE_BUSY) == REC_CODE_BUSY) {
        	/* printf1("CAN_int, tx REC_CODE_BUSY"); */
	};

/* check whether an RTR message was received in the transmit channel */
  if ((CAN_READ_CTRL(TRANSMIT_OBJ) & REC_CODE_FULL) == REC_CODE_FULL)
  {
            /* printf2("CAN_int, got receive interrupt on transmit channel,  OID= ", CAN_READ_OID(TRANSMIT_OBJ)); */

  	    if (pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
	              CO_LINE_PARA_ARRAY_INDEX .eStat == FULL)
	    {
	    	/* printf1("CAN_int, got receive interrupt on transmit channel, buffer full"); */
	    }
	    else
	    {
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .wCOB_ID =	CAN_READ_OID(TRANSMIT_OBJ);
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .bLength =	CAN_OBJ[TRANSMIT_OBJ].ctl_status & 0x0f;
		CO_MEMCPY(pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		    CO_LINE_PARA_ARRAY_INDEX .pData, &CAN_OBJ[TRANSMIT_OBJ].MSG0, 8);
		pRX_Buffer[bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX]
		      CO_LINE_PARA_ARRAY_INDEX .eStat = FULL;
		if (++bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX == CONFIG_RX_BUFFER_SIZE)
		{
		    bRX_WriteIndex CO_LINE_PARA_ARRAY_INDEX = 0;
		}
	    }
	    // keep the buffer locked to avoid race conditions betwen
	    // receive and transmit.

  } else {
//       printf2("CAN_int, got transmit interrupt : OID= ", CAN_READ_OID(TRANSMIT_OBJ));
  }


	eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
	GetNext_TX_Request(CO_LINE_PARA); /* get next transmission request */
   
	CAN_RESET_BIT(CAN_InterruptFlags, (1 << TRANSMIT_OBJ));
    }
/* --------------------------------------------------------------------------*/
    /* Reset RXOK, TXOK and LEC(Last Error Code) in CAN-SR */

#endif 	/* CONFIG_FULLCAN */

/* ------------ Both Full CAN and Basic CAN ---------------------------------*/
    (void) CAN_READ(CAN_TimerRegister);  /* unlock any message buffers */
    TRACE(0x0F);
    ENABLE_CAN_INTERRUPTS(CO_LINE_PARA);	/* enable CAN interrupt */
//    printf1("CAN_int, interrupts enabled");

//    RESET_BIT(1);
// POP_ALL_REGS();
}

/*******************************************************************
*
*++ GetNext_TX_Request - transmits message from the queue
*-- GetNext_TX_Requset - bertrgt eine Message aus dem Puffer
*
*++ Gets the next CAN object structure from the transmit buffer
*-- Holt die nchste CAN Objekt Struktur aus dem Transmit Puffer. 
*
*++ A Full CAN controller in Basic CAN mode
*++ uses Object TRANSMIT_OBJ to Transmit a COB.
*++ In Full CAN mode the object specified in bChannel
*++ of the COB structure is used.
*-- Ein Full CAN Controller im Basic CAN Mode
*-- benutzt das Objekt TRANSMIT_OBJ zum bertragen des COB.
*-- Im Full CAN Mode wird der objektspezifische Kanal
*-- aus bChannel der COB Struktur benutzt.
* 
* RETURNS
* .TP
*++ nothing
*-- nichts
*
*/

void GetNext_TX_Request(CO_LINE_PARA_DECL
	)
{
UNSIGNED8 bChannel;
UNSIGNED8 bLength;
UNSIGNED16 wID;

# ifdef CONFIG_MULT_LINES
    tou_can_t *tou_can = tou_can_addr CO_LINE_PARA_ARRAY_INDEX;
# endif

#if defined(CONFIG_FULLCAN) && !defined(CONFIG_ONLY_ONE_TRANSMIT_CHANNEL)
    bChannel = pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
               CO_LINE_PARA_ARRAY_INDEX .bChannel;
#else
    bChannel = TRANSMIT_OBJ;
#endif

    /* if CAN controller is busy return */
    if(eSending CO_LINE_PARA_ARRAY_INDEX == CO_TRUE)  {
    	/* EZ */
//    	printf2("GetNext_TX_Request: Still sending, bchannel = ", bChannel);
        return;
    }
//    printf2("GetNext_TX_Request for line ",canLine);
 

    if(pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eStat == FULL)
    {
    	/* -- eType ausgeben EZ */
//    	printf2("GetNext_TX_Request: actual etype = ",
//    	        pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
//		CO_LINE_PARA_ARRAY_INDEX .eType);
    	/* Transmit COB available */
    	/* Bit NOT_READY setzen */
#ifdef CONFIG_FULLCAN
	CAN_WRITE_CTRL(bChannel, TRANS_CODE_NOT_READY, 1);
#else
	CAN_WRITE_CTRL(TRANSMIT_OBJ, TRANS_CODE_NOT_READY, 1);
#endif

        wID = pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX] CO_LINE_PARA_ARRAY_INDEX .wCOB_ID;
//        printf2("GetNext_TX_Request, Cob_ID= ",wID);
	if ( pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == RTR_RX_COB) {
	    CAN_WRITE_OID_RTR(bChannel, wID);
//    	    printf1("GetNext_TX_Request: Setting RTR_bit");
	} else  {
	    CAN_WRITE_OID(bChannel, wID);
//    	    printf1("GetNext_TX_Request: not Setting RTR_bit");
	}
	/* data frame */
	if (   pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == TX_COB
	    || pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == RTR_TX_COB)
	{
            /* writing data to controller */
	    CO_MEMCPY ( (void *)&CAN_OBJ[bChannel].MSG0,
		     (void *)&pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
			CO_LINE_PARA_ARRAY_INDEX .pData[0],
		     pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
			CO_LINE_PARA_ARRAY_INDEX .bLength
		     );
	}
	bLength = pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .bLength;

	/* activate message transmission */
	/* reset CPUUPD and set Bits MSGVAL and TXRQ */


	/* the next statements should be atomar.
	 * that means that incrementing the pointer and settingthe flag
	 * must be done with sending the object.
	 */
	DISABLE_CPU_INTERRUPTS();
	eSending CO_LINE_PARA_ARRAY_INDEX = CO_TRUE;

#ifdef CONFIG_FULLCAN
	if (pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == TX_COB ||
	     pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == RTR_RX_COB)  {
	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONCE, bLength);
	}
	if (pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX]
		CO_LINE_PARA_ARRAY_INDEX .eType == RTR_TX_COB)  {
#if 0
	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONLY_RTR, bLength);
	    /* transmit RTR's will not be send yet */
	    eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
#else
//	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONCE_RTR_ALWAYS, bLength);
	    CAN_WRITE_CTRL(bChannel,TRANS_CODE_TRANSMIT_ONCE, bLength); 
	    eSending CO_LINE_PARA_ARRAY_INDEX = CO_FALSE;
#endif
	}
#else
	CAN_WRITE_CTRL(TRANSMIT_OBJ,TRANS_CODE_TRANSMIT_ONCE, bLength);
#endif
	pTX_Buffer[bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX ]
		CO_LINE_PARA_ARRAY_INDEX .eStat = EMPTY;

	/* increment the readpointer only if it is not equal writepointer */
	if (bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX 
		!= bTX_WriteIndex CO_LINE_PARA_ARRAY_INDEX)  {
	    if (++bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX==CONFIG_TX_BUFFER_SIZE)
	        bTX_ReadIndex CO_LINE_PARA_ARRAY_INDEX = 0;
	}
//	printf2("GetNext... ctl_status: ",CAN_OBJ[bChannel].ctl_status);
//	printf2("GetNext... std_id: ",CAN_OBJ[bChannel].std_id);
	ENABLE_CPU_INTERRUPTS();
    }
}


/*******************************************************************
*
*++ Transmit_COB - transmits a COB
*-- Transmit_COB - Senden eines COB
*
* NOMANUAL
*
* RX_COB	receive message
* RTR_RX_COB	receive message, which can be requested by the local device
* TX_COB	transmit message
* RTR_TX_COB	transmit message, which can be requested by remote devices
* 
* That means that both TX_COB and RTR_TX_COB are data messages to transmit.
* That other can be of type RTR_RX_COB,
* that means transmit a RTR message from local node.
*
* RETURNS
* .TP
* nothing
*
*/
void Transmit_COB(
     COB_T     *pCOB,   /* pointer to COB in list */
     UNSIGNED8 *pMsg    /* pointer to data */
     )
#ifdef TARGET_LINUX

#define CAN_RTR_SET			0x10
{
canmsg_t message;
int ret;
extern int debug;

    message.id = pCOB->wID;
    message.length = pCOB->bLength;;
    if(pCOB->eType == RTR_RX_COB) {
	message.length = pCOB->bLength & CAN_RTR_SET;
    }
    CO_MEMCPY(message.data, pMsg, pCOB->bLength);

    if(debug) {
        char *tmp_type;
        switch(pCOB->eType) {
	    case RX_COB: tmp_type  = "RX_COB";
		    break;
	    case TX_COB: tmp_type  = "TX_COB";
		    break;
	    case RTR_RX_COB: tmp_type  = "RTR_RX_COB";
		    break;
	    case RTR_TX_COB: tmp_type  = "RTR_TX_COB";
		    break;
	    default:
	    		tmp_type = "invalid";
        }

	PRINTF("Transmit ID %d, Type  %s\n", pCOB->wID, tmp_type);
    }
    ret = write(can_fd, &message, 1);
    if(ret == -1) {
	perror("write error; ");
    } else if (ret == 0) {
	PRINTF("Node %s: transmit time out\n", can_device);
    } else {
    ; /* sent ok */
    }
}
#else
{
#ifdef CONFIG_MULT_LINES
UNSIGNED8 canLine = pCOB->canLine;
#endif

    Insert_TX_Request(pCOB, pMsg);
    /* get next transmission request from queue */
    GetNext_TX_Request(CO_LINE_PARA);
}

#endif


/*______________________________________________________________________EOF_*/
/* EZ; where is ReceivedCAN used? 
int ReceivedCAN( COB_T *pCOB, UNSIGNED8 *pMsg)
{
UNSIGNED8 bLength;

    if(CAN_TEST_BIT(CAN_InterruptFlags, (1 << ALL_RECEIVE_OBJ))) {
	pCOB->wID = CAN_READ_OID(ALL_RECEIVE_OBJ);
	bLength = CAN_OBJ[ALL_RECEIVE_OBJ].ctl_status & 0xf;
	pCOB->bLength = bLength;
	CO_MEMCPY(pMsg, (void *)&CAN_OBJ[ALL_RECEIVE_OBJ].MSG0, bLength);

	CAN_RESET_BIT(CAN_InterruptFlags, (1 << ALL_RECEIVE_OBJ));
	CAN_WRITE_CTRL(ALL_RECEIVE_OBJ, REC_CODE_EMPTY, 8);
	return 1;
    }
    return 0;
}
*/

/*______________________________________________________________________EOF_*/
