/*
 * can_async - can4linux CAN driver module
 *
 * can4linux -- LINUX CAN device driver source
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * 
 * Copyright (c) 2001 port GmbH Halle/Saale
 * (c) 2001 Heinz-Jrgen Oertel (oe@port.de)
 *          Claus Schroeter (clausi@chemie.fu-berlin.de)
 * Copyright (c) 2003  Phil Wilshire ( philwil@sysdcs.com)

 *------------------------------------------------------------------
 * $Header: $
 *
 *--------------------------------------------------------------------------
 *
 *
 * modification history
 * --------------------
 * $Log: can_async.c,v $
 *
 *
 *--------------------------------------------------------------------------
 */


/**
* \file can_async.c
* \author Phil Wilshire SDCS (philwil@sysdcs.com)
* $Revision: 1.1 $
* $Date: 2003/09/01 18:03:41 $
*
*
*/


/* header of standard C - libraries */
/* #include <linux/module.h>			 */

/* header of common types */

/* shared common header */

/* header of project specific types */

/* project headers */
#include "can_defs.h"

/* local header */

/* constant definitions
---------------------------------------------------------------------------*/

/* local defined data types
---------------------------------------------------------------------------*/

/* list of external used functions, if not in headers
---------------------------------------------------------------------------*/

/* list of global defined functions
---------------------------------------------------------------------------*/

/* list of local defined functions
---------------------------------------------------------------------------*/

/* external variables
---------------------------------------------------------------------------*/

/* global variables
---------------------------------------------------------------------------*/

/* local defined variables
---------------------------------------------------------------------------*/
/* static char _rcsid[] = "$Id: can_async.c,v 1.5 2002/08/08 18:03:41 oe Exp $"; */


/***************************************************************************/
/**
*
* \brief int fcntl(int fd, const int F_SETFL, int flags | FASYNC);
* causes the async handler to be called
* \param fd result of the open call
* \param flags is the result of fcntl(fd, F_GETFL)

*
*
* The fcntl call is used to set the device flags.
* With the FASYNC flag set this mecanism is activated
*
* \returns
* return 0 
* or -1 if an error occurred (in which case, errno is set appropriately).
*
* \par ERRORS
* the following errors can occur
* \arg \c ENXIO  the file is a device special file
* and no corresponding device exists.
* \arg \c EINVAL illegal \b minor device number
* \arg \c EINVAL wrong IO-model format in /proc/sys/Can/IOmodel
* \arg \c EBUSY  IRQ for hardware is not available
* \arg \c EBUSY  I/O region for hardware is not available

*/


//static int can_fasync( __LDDK_FASYNC_PARAM ) PSW not working try this
static int can_fasync(int fd, struct file *file, int count)
{
can_data_t *dev = (can_data_t *)file->private_data;
int retval = 0;

    DBGin("can_fasync");
    {
      retval = fasync_helper(fd, file, count, &dev->fasyncptr);
      DBGprint(DBG_DATA,(" -- async  count %d  retval %d\n", 
			 count,retval));
      if ( retval > 0 )
	retval = 0;
    }
    DBGout();
    return retval;
}

