/* can_mcf5282funcs.c  - Motorola FlexCAN functions
 *
 * can4linux -- LINUX CAN device driver source
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * 
 * Copyright (c) 2003 port GmbH Halle/Saale
 * (c) 2003 Heinz-Jrgen Oertel (oe@port.de)
 * (c) 2003 Phil Wilshire (philwil@earthlink.net)
 *------------------------------------------------------------------
 * $Header: $
 *
 *--------------------------------------------------------------------------
 *
 *
 *
 *  TRAMSMIT_OBJ  - used to transmit messages, possible are:
 *  	standard, extended, standard RTR, extended RTR
 *  RECEIVE_RTR_OBJ
 *  RECEIVE_STD_OBJ
 *  RECEIVE_EXT_OBJ
 *
 *
 * modification history
 * --------------------
 * $Log: $
 *
 * Phil Wilshire added extensive 5282 handling mechanisms.
 * error handling
 * buffer tracking and handling
 * Multiple tx fifos
 * Multiple device opens each with their own rx fifo
 * rx fifo distribution
 * RTR transmit to receive buffer handling
 *   etc
 */
#include "can_defs.h"
#include <asm/delay.h>
#include <linux/interrupt.h>
#include <linux/timer.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
# error use the Coldfire FLexCAN only with Kernel > 2.4 
#endif

void Can_soft_reset(int board);


/* int	CAN_Open = 0; */

/* timing values */
static const BTR_TAB_TOUCAN_T can_btr_tab_toucan[] = {
 { 10, CAN_PRESDIV_10K, CAN_PROPSEG_10K, CAN_PSEG1_10K, CAN_PSEG2_10K },
 { 20, CAN_PRESDIV_20K, CAN_PROPSEG_20K, CAN_PSEG1_20K, CAN_PSEG2_20K },
 { 50, CAN_PRESDIV_50K, CAN_PROPSEG_50K, CAN_PSEG1_50K, CAN_PSEG2_50K },
 { 100, CAN_PRESDIV_100K, CAN_PROPSEG_100K, CAN_PSEG1_100K, CAN_PSEG2_100K },
 { 125, CAN_PRESDIV_125K, CAN_PROPSEG_125K, CAN_PSEG1_125K, CAN_PSEG2_125K },
 { 250, CAN_PRESDIV_250K, CAN_PROPSEG_250K, CAN_PSEG1_250K, CAN_PSEG2_250K },
 { 500, CAN_PRESDIV_500K, CAN_PROPSEG_500K, CAN_PSEG1_500K, CAN_PSEG2_500K },
 { 800, CAN_PRESDIV_800K, CAN_PROPSEG_800K, CAN_PSEG1_800K, CAN_PSEG2_800K },
 { 1000, CAN_PRESDIV_1000K, CAN_PROPSEG_1000K, CAN_PSEG1_1000K, CAN_PSEG2_1000K },
 {0, 0, 0, 0, 0}  /* last entry */
};



/* 
 * PSW Test Code No longer used
 * if extended error handling is needed we may re use this
 * Currently you need to either Close all devices or issue 
 * a Can Reset ioctl command to reset the device
 *
 * here is a description of the old code...
 *
 * PSW tasklet / timer function to reenable interrupts
 *
 * When we get max_error_cnt interrupts we disable the
 * error irqs. 
 * at the same time we optionally trigger the timer to reenable
 * them 
 * The timer function is added by a tasklet error
 */


//int can_error_irq_tot = 0;

//int can_error_max = 4;
//int can_error_max_soft = 64;

int can_error_irq = 0;

//
// function prototypes 
//
static int Can_Set_RxFifos ( int board,int irq_obj);
static void CAN_SendFifo(int board,int irq_obj) ;
static int Can_error_irq_on(int board);
static int Can_error_irq_off(int board);

static int Can_send_common( int board , canmsg_t *tx , int tx_obj );
static int Can_idle_detect(int board);
static int Can_txbuffer_wait(int board, int txobj);

//
// PSW auto tune for txbuffer wait not really used yet
// set this to 1 to disable auto tune
//
static int send_message_autotune = 1;


//static int error_timer_added=0;
//static int error_timer_delay = 10;  



#if 0

// PSW Code left in for possible later use.
static void error_timer_action(unsigned long dummy)
{
  int board = 0;  // PSW hack
  // int canmcr;
  //int estat;
  msg_fifo_t *TxFifo;
  unsigned long flags;

  TxFifo = &Tx_Buf[board];
  // we may take longer than a tick to process
  del_timer(&error_timer);// NOpe this did not work

  printk(" Error on timer triggered at %d \n",(int)jiffies);
  // Make sure that the Can Device is NOT busy
  // then turn errors back on

  Can_error_irq_on(board);


  if (TxFifo->active == 1 ) {
    printk(" PSW ERROR ON Send Fifo (disabled) \n");
    
    save_flags(flags);cli();
    //CAN_SendFifo(board,obj ??);
    /* leave critical section */
    restore_flags(flags);
  }
}

//
//
//
static void error_on(unsigned long dummy)
{

      printk(" Error on tasklet triggered at %d timer_added %d\n",
	     (int)jiffies,error_timer_added);
      if ( error_timer_added == 0 ) {
	error_timer_added = 1;
	error_timer.expires  = jiffies + error_timer_delay;
	error_timer.data     = 0;
	error_timer.function = (void *)&error_timer_action;
	add_timer(&error_timer);
      }
}

DECLARE_TASKLET(error_tasklet, error_on, 0);
#endif


/* Board reset
   means the following procedure:
  set Reset Request
  wait for Rest request is changing - used to see if board is available
  initialize board (with valuse from /proc/sys/Can)
    set output control register
    set timings
    set acceptance mask
*/


#ifdef DEBUG
int CAN_ShowStat (int board)
{
    if (dbgMask && (dbgMask & DBG_DATA)) {
    printk(" MCR 0x%x,", CANinw(board, canmcr));
    printk(" ESTAT 0x%x,", CANinw(board, estat));
    printk(" IFLAGS 0x%x,", CANinw(board, iflag));
    printk(" IMASK 0x%x,", CANinw(board, imask));
    printk("\n");
    }
    return 0;
}
#endif


/* can_GetStat - read back as much status information as possible 
*
* Because the CAN protocol itself describes different kind of information
* already, and the driver has some generic information
* (e.g about it's buffers)
* we can define a more or less hardware independent format.
*
*
* exception:
* ERROR WARNING LIMIT REGISTER (EWLR)
* The SJA1000 defines a EWLR, reaching this Error Warning Level
* an Error Warning interrupt can be generated.
* The default value (after hardware reset) is 96. In reset
* mode this register appears to the CPU as a read/write
* memory. In operating mode it is read only.
* Note, that a content change of the EWLR is only possible,
* if the reset mode was entered previously. An error status
* change (see status register; Table 14) and an error
* warning interrupt forced by the new register content will not
* occur until the reset mode is cancelled again.
*/

int can_GetStat(
	struct inode *inode,
	CanStatusPar_t *stat
	)
{
unsigned int board = MINOR(inode->i_rdev);	
msg_fifo_t *Fifo;
unsigned long flags;


    stat->type = CAN_TYPE_FlexCAN;

    stat->baud = Baud[board];
    /* printk(" STAT ST 0x%x\n", CANin(board, canstat)); */
    stat->status = CANinw(board, estat);
    /* not available in the FlexCAN, lets fill 127 here */
    stat->error_warning_limit = 127;
    stat->rx_errors = CANin(board, rxectr);
    stat->tx_errors = CANin(board, txectr);
    /* error code is not available, use estat again */
    stat->error_code= CANinw(board, estat);

    printk(" PSW Get Stat not modified yet for txFifo and rx fifos\n");
    /* Disable CAN Interrupts */
    /* !!!!!!!!!!!!!!!!!!!!! */
    save_flags(flags); cli();
    Fifo = &Rx_Buf[board];
    stat->rx_buffer_size = MAX_BUFSIZE;	/**< size of rx buffer  */
    /* number of messages */
    stat->rx_buffer_used =
    	(Fifo->head < Fifo->tail)
    	? (MAX_BUFSIZE - Fifo->tail + Fifo->head) : (Fifo->head - Fifo->tail);
    Fifo = &Tx_Bufx[board];
    stat->tx_buffer_size = MAX_BUFSIZE;	/* size of tx buffer  */
    /* number of messages */
    stat->tx_buffer_used = 
    	(Fifo->head < Fifo->tail)
    	? (MAX_BUFSIZE - Fifo->tail + Fifo->head) : (Fifo->head - Fifo->tail);
    /* Enable CAN Interrupts */
    /* !!!!!!!!!!!!!!!!!!!!! */
    restore_flags(flags);
    return 0;
}

//
// PSW this code is due to be deprecated in a later release
//
void Can_BuffMode_init(int board) 
{
  int i;
  for(i = 0; i < 16; i++) {
    BuffMode[i] = 0; 
  }
  //
  // preset some modes
  //
  BuffMode[0] = CAN_BUF_TX; 
  BuffMode[1] = CAN_BUF_RX; 

}
//
// PSW set a selected buffer as Rx and set up the OID
//
void Can_Set_Rx(int board,int rxobj,int oid, int flags) 
{
  CAN_WRITE_CTRL(rxobj, REC_CODE_NOT_ACTIVE, 0);
  if ( flags ) {
    CAN_WRITE_XOID(rxobj, oid); /* set IDE - extended bit to zero */
  } else {
    CAN_WRITE_OID(rxobj, oid);	/* set IDE - extended bit to zero */
  }
  //  CAN_SetMask  (board, AccCode[board], AccMask[board] );
  CAN_WRITE_CTRL(rxobj, REC_CODE_EMPTY, 8);
  Can_Buf[rxobj].status = CAN_BUF_RX;

  // PSW Clear the Rx buffer device id
  Can_Buf[rxobj].dev = 0;
  (" PSW Can_Set_Rx enabling RX on Chan %d \n",rxobj);   

}



/* 
 * CAN_ChipReset - performs the first initialization or re-iniatalization of the chip
 *
 *
 */
int CAN_ChipReset (int board)
{
int i;

    DBGin("CAN_ChipReset " __TIME__);
	
    /* 
     * Initialize Port AS PAR to have Can TX/RX signals enabled 
     */
#define MCF5282_GPIO_PASPAR  	(*(volatile u16 *)(void *)(0x40100056))
    MCF5282_GPIO_PASPAR = 0x0FF0;
    /* printk("I/O %p %04x\n", (volatile u16 *)(void *)(0x40100056), MCF5282_GPIO_PASPAR); */
	

    /*
     * go to INIT mode
     * Any configuration change/initialization requires that the TouCAN be
     * frozen by either asserting the HALT bit in the
     * module configuration register or by reset.
     * For Init_CAN() we choose reset.
     */
    // PSW some debug hooks here
    //PSW CANresetw(board, canmcr, CAN_MCR_FRZ);
    /* CANoutw(board, canmcr, 0x0080); */
    //printk(" PSW ChipReset SOFT RESET\n");

    CANsetw(board, canmcr, CAN_MCR_SOFT_RST);
    udelay(10);
    /* Test Reset Status */
    if(CANtestw(board, canmcr, CAN_MCR_SOFT_RST) != 0) {
	MOD_DEC_USE_COUNT;
	DBGout();return -1;
    }

    udelay(10);
    /* Initialize the transmit and receive pin modes in control register 0 */
    CANout(board, canctrl0, 
    	  CAN_CTRL0_DISABLE_BOFF_INT	/* disable Bus-Off Interrupt	*/
    	+ CAN_CTRL0_DISABLE_ERR_INT	/* disable Error Interrupt	*/
    	+ 0				/*  4 , logic RX level		*/
    	+ 0				/*  1 , logic TX level		*/
    	);
    /* CANresetw(board, canmcr, CAN_MCR_HALT); */


    CAN_SetTiming(board, Baud[board]);

    /*
     * Select the internal arbitration mode
     * LBUF in CANCTRL1
     * LBUF Lowest Buffer Transmitted First
     * The LBUF bit defines the transmit-first scheme.
     * 0 = Message buffer with lowest ID is transmitted first.
     * 1 = Lowest numbered buffer is transmitted first.
     *
     * should have no impact here, the driver is using only one 
     * TX object
     * PSW we have modified the number of tx objects
     *
     * !! change the order of the next two statements  !!
     */
     
    CANset(board, canctrl1, CAN_CTRL1_LBUF);
    CANreset(board, canctrl1, CAN_CTRL1_LBUF);

    /*
     * Initialize message buffers.
     * The control/status word of all message buffers are written
     * as an inactive receive message buffer.
     */
    // PSW we will set up buffers according to the 
    // spec in Can_Buff
    // All buffers are set up to REC_CODE_NOT_ACTIVE
    // at first , subsequent operations will
    // set different values

    for(i = 0; i < 16; i++) {
	CAN_WRITE_CTRL(i, REC_CODE_NOT_ACTIVE, 0);
    }

    /* create a transmit object, that can be used for all kinds of messages */

    /* some receive objects
     * - RECEIVE_OBJ to receive all standard frames
     * - ??
     */
    //printk(" PSW (1) enabling RX on Chan %d \n",rxobj);   
    //Can_Set_Rx(board, rxobj,0,0); //OID 0 flags 0
    CAN_SetMask  (board, AccCode[board], AccMask[board] );
    CAN_SetMask14  (board, AccCode14[board], AccMask14[board] );
    CAN_SetMask15  (board, AccCode15[board], AccMask15[board] );

    for(i = 0; i < 16; i++) {
      if ((BuffMode[i] & CAN_BUF_RX ) > 0 ) {
	Can_Set_Rx(board, i, 0, 0);
	//printk(" PSW (2) enabling RX on Chan %d \n",i);   
      }
    }

   //CAN_SetMask  (board, AccCode[board], AccMask[board] );
    /*
     * - Initialize IARB[3:0] to a non zero value in CANMCR
     *   ( This is done in CAN-ChipReset() )
     * - Set the required mask bits in the IMASK register (for all message
     *  buffer interrupts) in CANCTRL0 for bus off and error interrupts,
     *  and in CANMCR for WAKE interrupt.
     */

    /* dont't forget error int's ! */
    CANsetw(board, canmcr, 1); // ?? PSW 
    // PSW enable them all
    //CANoutw(board, imask, (1 << rxobj)+(1 << TRANSMIT_OBJ));
    CANoutw(board, imask, 0xFFFF);

/* printk("imask at %p \n", &((canregs_t *)can_base[board])->imask ); */

    /* CAN_register_dump(); */
    DBGout();
    return 0;
}

//
// PSW this may be deprected
// use ioctl CONF_SETRXBUFS instead;

/* change the buff mode  of the selected CAN channel */
int CAN_SetBuffMode (int board, int idx, int val)
{
    DBGin("CAN_SetBufMode");
    DBGprint(DBG_DATA, ("BufMode[%d]=%d", idx, val));
    BuffMode[idx] = val;
    DBGout();
    return 0;
}


/* change the bit timings of the selected CAN channel */
int CAN_SetTiming (int board, int baud)
{
/* int custom=0; */
BTR_TAB_TOUCAN_T * table = (BTR_TAB_TOUCAN_T*)can_btr_tab_toucan;

// printk("PSW CAN_SetTiming baud %d\n",baud);
    DBGin("CAN_SetTiming");
    DBGprint(DBG_DATA, ("baud[%d]=%d", board, baud));
    /* enable changing of bit timings */
    CANsetw(board, canmcr, CAN_MCR_HALT);
    /* search for data from table */
    while(1) {
        if (table->rate == 0 || table->rate == baud) {
    	    break;
    	}	
    	table++;
    }
    if (table->rate == 0) {
	/* try to use baud  as custom specific bit rate 
	 * not implemented yet
	 */
	return -ENXIO;
    }

    /*
     * Set Timing Register values.
     * Initialize the bit timing parameters PROPSEG, PSEG1, PSEG2 and RJW
     * in control registers 1 and 2.
     *
     * The TouCAN module uses three 8-bit registers to set-up
     * the bit timing parameters required by the CAN protocol.
     * Control registers 1 and 2 contain the PROPSEG, PSEG1, PSEG2
     * and RJW fields which allow the user to configure
     * the bit timing parameters.
     * The prescaler divide register (PRESDIV) allows the user to select
     * the ratio used to derive the S-Clock from the system clock.
     * The time quanta clock operates at the S-clock frequency. 
     */
    CANout(board, presdiv, table->presdiv);
    CANout(board, canctrl2, ((table->pseg1 << 3)+ table->pseg2));
    CANout(board, canctrl1, (
    	  0		/* SAMP = 0 , 0/0x80		*/	
    	+ 0		/* LOOP = 0 , 0/0x40		*/	
    	+ 0		/* TSYNC= 0 , 0/0x20		*/
    	+ 0		/* LBUF = 0 , 0/0x10		*/
    	+ 0		/* RSVD = 0 , 0/0x08		*/
    	+ table->propseg)
    	);

    /* 
     * Stay in configuration mode; a call to Start-CAN() is necessary to
     * activate the CAN controller with the new bit rate
     */
    DBGprint(DBG_DATA,("canctrl2=0x%x presdiv=0x%x",
    		CANin(board, canctrl2), CANin(board, presdiv)) );

  
    Can_state[board] = 0;
    Can_soft_reset(board);
    //printk("PSW reset baud turning errors back on \n"); 
    Can_error_irq_on(board);

    DBGout();
    return 0;
}


int CAN_StartChip (int board)
{
/* int i; */
    RxErr[board] = TxErr[board] = 0L;
    DBGin("CAN_StartChip");


    /* clear interrupts */
    CANinw(board, iflag);
    CANoutw(board, iflag, 0xffff); /* 5282 overwrites with '1' */

    //printk(" PSW not resetting FRZ we want to halt \n");
    CANresetw(board, canmcr, CAN_MCR_HALT );

    //CANresetw(board, canmcr, CAN_MCR_FRZ);
    CANsetw(board, canmcr, CAN_MCR_FRZ); // PSW
    /* Interrupts on Rx, TX, any Status change and data overrun */
    /* Initialize the transmit and receive pin modes in control register 0 */
    CANset(board, canctrl0, 
    	  CAN_CTRL0_ENABLE_BOFF_INT	/* enable Bus-Off Interrupt	*/
    	+ CAN_CTRL0_ENABLE_ERR_INT	/* enable Error Interrupt	*/
    	);


    /* to do */
    /* CAN_register_dump(); */
    
    DBGout();
    return 0;
}


/* Disable all CAN activities */
int CAN_StopChip (int board)
{
  // PSW perhaps does not stop TX RX irqs because the HALT bit is ignored
    DBGin("CAN_StopChip");
    CANset(board, canmcr, CAN_MCR_HALT);
    CANset(board, canctrl0, 
    	  CAN_CTRL0_DISABLE_BOFF_INT	/* disable Bus-Off Interrupt	*/
    	+ CAN_CTRL0_DISABLE_ERR_INT	/* disable Error Interrupt	*/
    	);

    // disable all Rx / Tx interrupts here too
    //
    //printk(" PSW CAN_StopChip ...imask reset \n");
    //CANoutw(board, imask, 0x0000);

    DBGout();
    return 0;
}

/* set value of the output control register */
int CAN_SetOMode (int board, int arg)
{

    DBGin("CAN_SetOMode");
    DBGprint(DBG_DATA,("[%d] outc=0x%x", board, arg));

    DBGout();
    return 0;
}


/* FlexCAN only knows a 'mask' value, code is ignored */
int CAN_SetMask (int board, unsigned int code, unsigned int mask)
{

    DBGin("CAN_SetMask");
    DBGprint(DBG_DATA,("[%d] acc=0x%x mask=0x%x",  board, code, mask));
    CANoutw(board, rxgmskhi, mask >> 16);
    CANoutw(board, rxgmsklo, mask && 16);

    DBGout();
    return 0;
}
/* FlexCAN only knows a 'mask' value, code is ignored */
int CAN_SetMask14 (int board, unsigned int code, unsigned int mask)
{

    DBGin("CAN_SetMask14");
    DBGprint(DBG_DATA,("[%d] acc=0x%x mask=0x%x",  board, code, mask));
    CANoutw(board, rx14mskhi, mask >> 16);
    CANoutw(board, rx14msklo, mask && 16);

    DBGout();
    return 0;
}

/* FlexCAN only knows a 'mask' value, code is ignored */
int CAN_SetMask15 (int board, unsigned int code, unsigned int mask)
{

    DBGin("CAN_SetMask15");
    DBGprint(DBG_DATA,("[%d] acc=0x%x mask=0x%x",  board, code, mask));
    CANoutw(board, rx15mskhi, mask >> 16);
    CANoutw(board, rx15msklo, mask && 16);

    DBGout();
    return 0;
}


/* FlexCAN only knows a 'mask' value, code is ignored */
int CAN_SetEBuffMask (int board, unsigned int buff, 
		      unsigned int oid, 
		      int flags)
{

  int timer;

  DBGin("CAN_SetEBuffMask");
  
  DBGprint(DBG_DATA,("[%d] buff=%d oid=0x%x flags %d",  
		       board, buff, oid, flags));

  if ( buff > 15 ) return -1;
  if ( buff < 0 ) return -1;
  

  if ( Can_Buf[buff].status == CAN_BUF_TX ) {
    printk(" PSW yanking a TX buffer for RX \n");
    }
  printk(" PSW setting up Rx buffer %d OID %x \n",buff,oid);
  Can_Buf[buff].status = CAN_BUF_RX;
  
  CAN_WRITE_CTRL(buff, REC_CODE_NOT_ACTIVE, 0);
  
  if ( flags ) {
    CAN_WRITE_XOID(buff, oid); /* set IDE - extended bit to zero */
  } else {
    CAN_WRITE_OID(buff, oid);	/* set IDE - extended bit to zero */
  }
  
  CAN_WRITE_CTRL(buff, REC_CODE_EMPTY, 8);
  // PSW release all locks
  timer = CANinw(board, timer);
  
  DBGout();
  return 0;
}

int CAN_SetBuffMask (int board, unsigned int buff, unsigned int oid)
{
  return CAN_SetEBuffMask (board, buff, oid, 0);

}
//
// extended version
//
int CAN_SetBuffMaskE (int board, unsigned int buff, unsigned int oid)
{
  return CAN_SetEBuffMask (board, buff, oid, 1);

}

// PSW Special
// get the txmask from the device specified in the id
// look through all the possible buffers ( specified in the mask )
// status = 0 or Status = TX and not busy
//  
// example call tx_obj = CAN_findtxbuf(dev->id);
// PSW Deprecated this is now in the message 
#if 0
int CAN_findtxbuf(int id ) 
{
  int tbuffers;
  can_data_t * dev = &Can_Data[id];
  int i;
  
  tbuffers = dev->tbuffers;
  //printk(" findtxbuf id %d tbuffers %x \n",id,tbuffers);

  for ( i = 0 ; i < 16 ; i++ ) {
 
    //printk(" findtxbuf Can_Buf[%d].status = 0x%x  \n",i,Can_Buf[i].status);

    if ((tbuffers & (1 << i) ) == 0 ) {
      //printk(" findtxbuf skipping %d .. not allowed \n",i);
      continue;
    }
    

    if ( Can_Buf[i].status == 0 ) {
      //printk(" findtxbuf Found %d .. status = 0 \n",i);
      break;
    } 
    
    if ( Can_Buf[i].status == CAN_BUF_TX ) {
      // TX_BUSY is set if this buffer is not available
      //printk(" findtxbuf Found %d .. status = TX \n",i);
      break;
    } 

  }
  if ( i == 16 ) i = -1;
  return i;
}

#endif

//
// PSW used instead of CAN_READ_CTRL 
// Gets the buffer status without locking
//
int Can_status_nolock( int board, int obj ) 
{
  int stat;
  int timer;
  stat = CAN_READ_CTRLX(obj);
  // now read the timer to free the regs
  timer = CANinw(board, timer);
  return stat;

}

//
// PSW used instead of CAN_READ_CTRL 
// Gets the buffer status WITH locking
//
int Can_status_lock( int board, int obj ) 
{
  int stat;
  stat = CAN_READ_CTRLX(obj);
  return stat;

}
//
// PSW detects and waits for a non idle status on the bus
// we do not use this at present
//
static int Can_idle_detect(int board)
{
  int stat;
  /* 
     IDLE - Idle Status. The IDLE bit indicates, when there is activity
     on the CAN bus
     1 - The CAn bus is Idle
     
     TX/RX - Transmission/receive status. Indicates when the FlexCAN module
     is transmitting or receiving a message. TX/RX has no meaning,
     when IDLE = 1
     0 - FlexCAN is receiving when IDLE = 0
     1 - FlexCAN is transmitting when IDLE = 0
  */

  // PSW TODO we can still find and queue a tx buffer 
  // We only need to do this is the TRANSMIT_OBJ buffer is busy
  // 
  while ( 
	 ((stat = CANinw(board, estat)) & (CAN_ESTAT_IDLE + CAN_ESTAT_TX_RX))
	 == CAN_ESTAT_TX_RX
	 ) {
    
    if( current->need_resched ) schedule();
    /* if( need_resched ) schedule(); */
  }
  return 0;
}

//
// PSW wait for a TX buffer to become free
//
// If this buffer is an RX buffer it will timeout
// If it is an RTR Buffer then it will override it.
//
//
static int Can_txbuffer_wait(int board, int txobj) {
  int stat;
  int txcount;
  int done;
  done = 0;
  txcount = 0;
  while ( txcount < send_message_autotune ) {
    stat = (Can_status_nolock(board,txobj) >> 4) & 0x000F; 

    switch (stat) {
    case TRANS_CODE_NOT_READY: /* 8 */
    case REC_CODE_NOT_ACTIVE:  /*0x0*/
      done = 1;
      break;
    case TRANS_CODE_TRANSMIT_ONLY_RTR: /*0xa*/
    case TRANS_CODE_TRANSMIT_ONCE_RTR_ALWAYS: /*0xe*/
      printk(" PSW RTR stat %x overridden ....\n",stat);
      done = 1;
      break; // RTR Always

    case TRANS_CODE_TRANSMIT_ONCE: /*0xc*/
      break; // Busy just wait


    case REC_CODE_EMPTY:   /*0x4*/ 
    case REC_CODE_FULL:    /*0x2*/ 
    case REC_CODE_OVERRUN: /*0x6*/ 
    case REC_CODE_BUSY:    /*0x1*/ 
      //      if ( txcount == 0 ) 
      //printk(" PSW RX buffer used for Tx  ....\n");
      // not OK to use it but this could be an error
      // this is probably an  rtr tx buffer
      return 1;

      break;
    }

    if ( done == 1 ) break;

    txcount++;
    udelay(5);
    
  }
  
  
  //
  // Some auto tune handling
  // disable if send_message_autotune == 1

  if ( txcount == send_message_autotune ) {
    if ((send_message_autotune > 1 ) && 
	(send_message_autotune < 20 ) ){
      send_message_autotune++;
      printk(" CAN_SendMessage tune now %d\n", 
	     send_message_autotune);
    }
    return 1;  // error
  }
  
  // allow tune down
  if ( (send_message_autotune > 1 ) && 
       (txcount < send_message_autotune/5 )) {
    send_message_autotune--;
    printk(" CAN_SendMessage tune now %d\n", 
	   send_message_autotune);
    
  }
  return 0;
}
//
// PSW special 
// Populates the Can_buf.dev structure
//
int CAN_SendMessageX (int board, canmsg_t *tx,void * xdev)
{
  can_data_t * dev = (can_data_t*)xdev;
  //  u16 stat;
  u16 ctrl;
  int txobj;

  DBGin("CAN_SendMessageX");

  txobj = tx->buffer;

#if 0
  Can_idle_detect(board);
#else
  //
  // wait for the txbuffer to become empty
  //
  if (Can_txbuffer_wait(board, txobj)) {
    ctrl = Can_status_nolock(board,txobj);
    //printk(" PSW CSX buffer not empty ctrl %x ...\n",ctrl);
    DBGout();
    return 0;
  }
#endif
  
  Can_send_common(board, tx , txobj);
  // Just in case we locked it
  ctrl = Can_status_nolock(board,txobj);

  // PSW record who sent the data and set buf busy
  Can_Buf[txobj].dev = dev;

  DBGout();
  return 1; // PSW hmm is this right ??
}


//
// ioctl send messsage function
//
int CAN_SendMessage (int board, canmsg_t *tx)
{
  int ret;
int txobj;  // PSW
    txobj = tx->buffer;
    DBGin("CAN_SendMessage");

#if 0
    Can_idle_detect(board);
#else
    if (Can_txbuffer_wait(board, txobj))
      return 0;
#endif

    /* DBGprint(DBG_DATA,( */
    /* "CAN[%d]: tx.flags=%d tx.id=0x%lx tx.length=%d stat=0x%x", */
    /* board, tx->flags, tx->id, tx->length, stat)); */
    ret = Can_send_common(board, tx, txobj);

    DBGout();
    return ret;
}



/* look if one of the receive message objects has something received */
int CAN_GetMessage (int board, canmsg_t *rx )
{
int stat;
int i = 0;
int ctrl;
int rxobj = RECEIVE_OBJ;
int irqsrc2; // PSW for iflag reread
  
 DBGin("CAN_GetMessage");
 printk(" PSW Can GetMessage should not be used as is\n");
 stat = CANinw(board, estat);
 DBGprint(DBG_DATA,("0x%x: stat=0x%x iflag=0x%x imask=0x%x" ,
		    Base[board], stat,
		    CANinw(board, iflag),
		    CANinw(board, imask)));

 rx->flags  = 0;
 rx->length = 0;

 /* CAN_register_dump(); */
 /* CAN_object_dump(RECEIVE_OBJ); */
    i = CANinw(board, iflag);
    if( i & (1 << rxobj)) {
	/* reading the control status word of the identified message
	 * buffer triggers a lock for that buffer.
	 * A new received message frame which maches the message buffer
	 * can not be written into this buffer while it is locked
	 */

        while (((ctrl = Can_status_nolock(board,rxobj) ) & 
		(REC_CODE_BUSY << 4)) == (REC_CODE_BUSY << 4)) {
	    /* 20 cycles maximum wait */
	    /* printf1("CAN_int, rx REC_CODE_BUSY"); */
	}
	/* printk("message received 0x%x\n", ctrl); */
	rx->length = ctrl & 0x0f;
	rx->id =  CAN_READ_OID(rxobj);
	memcpy((char *)&rx->data[0], (char *)&(CAN_OBJ[rxobj].MSG0), 8);

	/* clear interrupts */
	/* Int is cleared when the CPU reads the interrupt flag register iflag
	 * while the associated bit is set, and then writes it back as '1'
	 * (and no new event of the same type occurs
	 * between the read and write action.)
	 * ! This is opposite to the TouCAN module, where the iflag bit
	 * has to be written back with '0'
	 */

	// PSW no ..this may not work
	// CANsetw(board, iflag, (1 << rxobj));

	irqsrc2 = CANinw(board, iflag); // PSW reread the iflag
	CANoutw(board, iflag,(1 << rxobj) );

	/* Reset message object */
	CAN_WRITE_CTRL(rxobj, REC_CODE_EMPTY, 8);
	/* reading the free running timer will unlock any message buffers */
	(void) CANinw(board, timer);
	i = 1; /* signal one received message */
    } else {
	i = 0;
    }

    DBGout();
    return i;

}

int CAN_VendorInit (int minor)
{
    DBGin("CAN_VendorInit");
    //    init_timer(&error_timer);


/* 1. Vendor specific part ------------------------------------------------ */
    can_range[minor] = 0x180;

/* End: 1. Vendor specific part ------------------------------------------- */

    /* Request the controllers address space */

    /* looks like not needed in uClinux with internal ressources ? */
    /* It's Memory I/O */
    if(NULL == request_mem_region(Base[minor], can_range[minor], "CAN-IO")) {
	return -EBUSY;
    }

    /* not necessary in uClinux, but ... */
    can_base[minor] = ioremap(Base[minor], can_range[minor]);
    /* now the virtual address can be used for the register address macros */

/* 2. Vendor specific part ------------------------------------------------ */



/* End: 2. Vendor specific part ------------------------------------------- */

    if( IRQ[minor] > 0 ) {
        if( Can_RequestIrq( minor, IRQ[minor] , CAN_Interrupt) ) {
	     printk("Can[%d]: Can't request IRQ %d \n", minor, IRQ[minor]);
	     DBGout(); return -EBUSY;
        }
    }
    DBGout(); 
    return 1;

}

//
// PSW Set a fifo with Rx Data
// locks the buffer and leaves it locked .. Sorry
//  use Can_status_nolock( board,rx_obj )  to unlock it
//

int Can_Set_Fifo(int board, msg_fifo_t * Fifo, int rx_obj, int ixd) 
{
  int ctrl;
  int dummy;
  ctrl = Can_status_lock( board,rx_obj ) ; // PSW lock it 


  do_gettimeofday(&(Fifo->data[Fifo->head]).timestamp);
  /* preset flags */
  (Fifo->data[Fifo->head]).flags =
    (Fifo->status & BUF_OVERRUN ? MSG_BOVR : 0);

  (Fifo->data[Fifo->head]).id =  CAN_READ_OID(rx_obj);
  
  /* get message length,  strip length code */ 
  dummy = ctrl & 0x0f;
  (Fifo->data[Fifo->head]).length = dummy;
  dummy %= 9;	/* limit count to 8 bytes */
  (Fifo->data[Fifo->head]).buffer = rx_obj;

  memcpy((char *)&(Fifo->data[Fifo->head]).data[0], 
	 (char *)&(CAN_OBJ[rx_obj].MSG0), dummy);

  Fifo->status = BUF_OK;
  Fifo->head = ++(Fifo->head) % MAX_BUFSIZE;
  
  if(Fifo->head == Fifo->tail) {
    printk("CAN Buff [%d] RX Fifo_num(%d): FIFO overrun\n", rx_obj,ixd);
    Fifo->status = BUF_OVERRUN;
  } 

  return 0;
}

//
// PSW used after a soft reset
// resets the HALT bit
//
int Can_resume(int board)
{ 
  //int canmcr;
  CANresetw(board, canmcr, CAN_MCR_HALT );
  udelay(10);
  //canmcr = CANinw(board, canmcr);
  //printk(" PSW Can resume canmcr %x \n",canmcr);
  return 0;
}

//
// PSW attepmts to halt the chip and looks for a halt response
//
int Can_halt(int board)
{
  //int		estat;
  int		canmcr;
  int ret;
  int loop;
  int max_loop;

  loop = 0;
  max_loop = 10;
  ret = 0;
  //estat = CANinw(board, estat);
  //printk(" PSW ERROR ON canmcr %x estat %x \n",canmcr,estat);
  CANsetw(board, canmcr, CAN_MCR_HALT );

  udelay(10);

  while((loop < max_loop) && (CANtestw(board, canmcr, CAN_MCR_FRZ_ACK) == 0)) {
    canmcr = CANinw(board, canmcr);
    printk(" PSW note  no FRZ_ACK in %d uSecs mcr = %x\n" , 
	   loop * 10,canmcr);
    udelay(10);

    loop ++;
  }
  if (loop >= max_loop ) ret = -1; 
  return ret;
}

//
// PSW debug used to show us what is going on
// Show irq flags and masks
// The id is set to an id string for debug
//
void Can_show_i(char * id) 
{

  volatile unsigned short * iflg;
  volatile unsigned short * imsk;
  volatile unsigned short * estat;
  volatile unsigned short * irmskh;
  volatile unsigned short * irmskl;
      
  estat  = (volatile unsigned short *) (MCF_MBAR + 0x1c0020 );
  imsk  = (volatile unsigned short *) (MCF_MBAR + 0x1c0022 );
  iflg = (volatile unsigned short *) (MCF_MBAR + 0x1c0024 );
  irmskh  = (volatile unsigned short *) (MCF_MBAR + 0x1c0010 );
  irmskl  = (volatile unsigned short *) (MCF_MBAR + 0x1c0012 );
  
  if (id ) {
    printk(" PSW *(%s)* iflg %x  imsk %x estat %x",id,*iflg, *imsk,*estat);
    printk(" irmh %x irml %x\n",*irmskh,*irmskl);
  } else {
    printk(" PSW *(***)* iflg %x  imsk %x estat %x",*iflg, *imsk,*estat);  
    printk(" irmh %x irml %x\n",*irmskh,*irmskl);
  }
}
//
// PSW some debug code
// Show irq flags and set mask
//
void Can_set_i(char * id) 
{

  volatile unsigned short * iflg;
  volatile unsigned short * imsk;
  volatile unsigned short * irmskh;
  volatile unsigned short * irmskl;
  //volatile unsigned short * estat;
  unsigned short old_imsk;
  int idb;
  int board=0;
  int ctrl;

  //estat  = (volatile unsigned short *) (MCF_MBAR + 0x1c0020 );
  imsk  = (volatile unsigned short *) (MCF_MBAR + 0x1c0022 );
  iflg = (volatile unsigned short *) (MCF_MBAR + 0x1c0024 );
  irmskh  = (volatile unsigned short *) (MCF_MBAR + 0x1c0010 );
  irmskl  = (volatile unsigned short *) (MCF_MBAR + 0x1c0012 );
  
  old_imsk= *imsk;
  if ( *imsk != 0xffff ) *imsk=0xffff;


  printk(" PSW *(%s)* iflg %x  imsk %x old imsk %x",id,
	 *iflg, *imsk,old_imsk);
  printk(" irmh %x irml %x\n",*irmskh,*irmskl);

  printk(" PSW ctrl ");
  for ( idb = 0 ; idb < 16 ; idb++ ) {
    ctrl = Can_status_nolock(board,idb) ;
    if (idb == 8 )   printk("\n PSW ctrl ");
    printk(" %d(%x) ",idb,ctrl);
  }
  printk("\n");
}
//
// PSW error handling
//  turns the error IRQ off to stop a system lock up
//
static int  Can_error_irq_off(int board) 
{
  int estat;
  estat = Can_halt(board);
  
  CANreset(board, canctrl0, 
	   CAN_CTRL0_ENABLE_ERR_INT   /* disable Error Irq	*/
	   );
  
  Can_resume(board); 
  
  return 0;
}

//
// PSW turns errors back on
//
static int Can_error_irq_on(int board)
{
  //int loop;
  //int canmcr;
  //int estat;

  // wait for FRZ_ACK
  Can_halt(board);

   
  //printk(" PSW ERROR ON FRZ_ACK OK resetting error IRQ\n");
  CANset(board, canctrl0, 
	 CAN_CTRL0_ENABLE_ERR_INT   /* enable Error Irq	*/
	 );
  
  CANresetw(board, canmcr, CAN_MCR_HALT );
  udelay(10);
  
  //canmcr = CANinw(board, canmcr);
  //printk(" PSW ERROR ON canmcr again %x \n",canmcr);
  return 0;
}

//
// PSW soft reset
// recovers masks after a soft reset
//
void Can_soft_reset(int board)
{
  //int ic;
  int imask;
  int irmskh;
  int irmskl;
  volatile unsigned short * imaskp;
  volatile unsigned short * irmskhp;
  volatile unsigned short * irmsklp;
  imaskp  = (volatile unsigned short *) (MCF_MBAR + 0x1c0022 );
  irmskhp  = (volatile unsigned short *) (MCF_MBAR + 0x1c0010 );
  irmsklp  = (volatile unsigned short *) (MCF_MBAR + 0x1c0012 );

  //Can_show_i("before Soft reset");
  imask = *imaskp;
  irmskh = *irmskhp;
  irmskl = *irmsklp;
  
  CANsetw(board, canmcr, CAN_MCR_SOFT_RST); // Soft reset the board

  //Can_show_i("after soft reset");
  *imaskp = imask;
  CAN_SetMask  (board, AccCode[board], AccMask[board] );
  CAN_SetMask14  (board, AccCode14[board], AccMask14[board] );
  CAN_SetMask15  (board, AccCode15[board], AccMask15[board] );
  
  
  //Can_show_i("after mask restore");

  //	  for (ic = 0 ; ic < 10 ; ic ++ ) {
  // estat = CANinw(board, estat);
  // canmcr = CANinw(board, canmcr);
  // printk(" PSW(%d) SOFT reset canmcr %x estat %x iswp %x\n",
  //	   ic, canmcr,estat,*iestatp);
  //  udelay(20);
  //  if (ic == 0 ) 

  Can_resume(board); // take out of HALT mode

  //}
}


//
// PSW for each message we receive. send to all the listeners
// each listener will flag the buffers it is interested in
// by setting the rbuffers flag
// 
static int Can_Set_RxFifos ( int board,int rx_obj)
{
  int ixd;
  can_data_t *  dev;

  // For each Can_Data that enables have rx_obj add the data
  
  DBGprint(DBG_DATA, (" Setting Fifo Data\n")); 
  for ( ixd = 0 ; ixd < 16 ; ixd++ ) {
    if (Can_Data[ixd].id == -1 ) continue;
    if ((Can_Data[ixd].rbuffers & (1 << rx_obj))> 0 ) { 
      Can_Set_Fifo(board,&Can_Data[ixd].RxFifo,rx_obj,ixd);
      
      // kick the select call PSW
      wake_up_interruptible(&Can_Data[ixd].RxWait); 
      
      // kick the async handler for the process
      dev = (can_data_t * )&Can_Data[ixd];
      if (dev)
	kill_fasync(&dev->fasyncptr, SIGIO, POLL_IN);
    }
  }
  Can_status_nolock( board,rx_obj ) ; // PSW unlock it
  return 0;
}


/*
 * The plain CAN interrupt
 *
 *				RX ISR           TX ISR
 *                              8/0 byte
 *                               _____            _   ___
 *                         _____|     |____   ___| |_|   |__
 *---------------------------------------------------------------------------
 * 1) Motorola ColdFire 5282
 *  66 MHz, xxxxxx bogomips
 *                              __/__s            __ s
 * 
 * 1) 1Byte = (42-27)/8 =      s
 * 2) 1Byte = (24-12)/8 =      s
 *
 *
 *
 * RX Int with two Receive channels:
 * 1)                _____   ___
 *             _____|     |_|   |__
 *                   30    5  20  s
 *   first ISR normal length,
 *   time between the two ISR -- short
 *   sec. ISR shorter than first, why? it's the same message
 */

void CAN_Interrupt ( int irq, void *dev_id, struct pt_regs *ptregs )
{
int		board;
int		estat;
int		canmcr;
unsigned long	flags;
int		ctrl;
 int 		irqsrc;
 int 		irqsrc2; // PSW for iflag reread
 //msg_fifo_t	*RxFifo; 
msg_fifo_t	*TxFifo;
//int		last_tx_id;

//int i;
int irq_obj;  // PSW allow a flexible buffer selection
int tx_wait_delay; // PSW stops infinite wait on send

can_data_t * dev; // PSW used for fasync

/* do we have some LEDS on the ColdFire EVA ?? */

/* printk(" CAN ISR %d\n", irq);  */


    board = *(int *)dev_id;

    // PSW deprecated 
    //RxFifo = &Rx_Buf[board]; 
    //TxFifo = &Tx_Buf[board];



    /* One of the following can actually occur:
       BusOff
       ErrInt - in this case:
       	bits 4-5 - FCS - Fault confinment state tells us
       	 00 - all ok, error active mode
       	 01 - error passive mode
        bit 8 - RX warning level reached
        bit 9 - Tx warning level reached

       At the moment we aren't interested in more information

       all bits in estat are read only except for
       BOFF_INT WAKE_INT und ERR_INT which are interrupt sources 
       and can be written by the host to '0' to reset Interrupt
       conditions.
    */
    estat = CANinw(board, estat);
#if 0
if (irq == 152) {
CAN_register_dump();
	CANresetw(board, estat, 7);
CAN_register_dump();
CAN_object_dump(0);
}
#endif
//
// PSW use a small state machine to process this
// we turn off the irq
    if ((irq == 152) && (estat & (CAN_ESTAT_ERR_INT)))  {
      can_error_irq++;
      printk(" PSW ERROR state %d irq %d at %d\n",
	     Can_state[board],irq, (int) jiffies);

      switch(Can_state[board]) {

      case 0:
	Can_state[board] = 1;
	// PSW  try a soft_ reset
	Can_soft_reset(board);
	break;
      case 1:
	Can_state[board] = 2;
	// PSW turn error irq off
	Can_error_irq_off(board);
      default:
	Can_state[board]++;
	printk(" PSW ERROR irq with irqs OFF ??? count %d \n",
	       Can_state[board]);
	Can_error_irq_off(board);
	break;
      }
      return;
    }

    // PSW fall through to this for Wake and BOFF
    if ((irq > 151 ) && (estat &
       (CAN_ESTAT_BOFF_INT + CAN_ESTAT_ERR_INT + CAN_ESTAT_WAKE_INT)))  {

      volatile unsigned short * iestatp;
      volatile unsigned short * imcrp;

      iestatp  = (volatile unsigned short *) (MCF_MBAR + 0x1c0020 );
      imcrp = (volatile unsigned short *) (MCF_MBAR + 0x1c0000 );

	/* we have an error interrupt
	 * later on we can move this error handling at the end of the ISR
	 * to have better RX response times */
	/* printk(" error status 0x%04x \n", estat); */
	/* reset all error conditions, we have saved them in estat 
	 * BusOff disables the Interrupt generation itself by resetting
	 * the mask in canctrl0.
	 * ErrInt - to clear this bit, first read it (already done)
	 * then write as a zero.
	 *
	 * we do reset all possible interrupts and look what we got later
	 */

        estat = CANinw(board, estat);
        canmcr = CANinw(board, canmcr);
	printk(" PSW irq %d after reset error canmcr %x estat %x iswp %x\n",
	       irq, canmcr,estat,*iestatp);

	/* check for Bus Off condition */
	if (estat & CAN_ESTAT_BOFF_INT) {
	    CAN_StopChip (board);
	    printk("We are BusOff now\n");
	}

#if 0
        // PSW move this to the error handler perhaps ??
	if (CAN_ESTAT_TX_WARN == (estat & (CAN_ESTAT_ERR_INT + CAN_ESTAT_TX_WARN))) {
	    printk("Tx Warning level reached\n");
	}
	if (CAN_ESTAT_RX_WARN == (estat & (CAN_ESTAT_ERR_INT + CAN_ESTAT_RX_WARN))) {
	    printk("Rx Warning level reached\n");
	}
	if (estat & (CAN_ESTAT_ERR_INT + CAN_ESTAT_FCS)) {
	    if(0 == (estat & CAN_ESTAT_FCS)) {
		printk("Going error active\n");
	    }
	    if(CAN_ESTAT_FCS0 == (estat & CAN_ESTAT_FCS)) {
		printk("Going error passive\n");
	    }
	}
#endif
    }

    /* check for message object interrupts */ 
    irqsrc = CANinw(board, iflag);
    //printk(" PSW 1 IRQ %d source %x\n",irq,irqsrc);

    if(irqsrc == 0) {
         /* first call to ISR, it's not for me ! */
         printk(" PSW IRQ with no source ...\n");
 
	 goto IRQdone_doneNothing;
    }

    do {
      /* loop as long as the CAN controller shows interrupts */
      
      // PSW  moved to the per process buffer system 
      
      //do_gettimeofday(&(RxFifo->data[RxFifo->head]).timestamp);
      
      /* preset flags */
      //(RxFifo->data[RxFifo->head]).flags =
      //(RxFifo->status & BUF_OVERRUN ? MSG_BOVR : 0);
      
      // PSW test all buffers
      for ( irq_obj = 0; irq_obj < 16 ; irq_obj++ ) { 
        //unsigned short  test;
        //int tcount = 3;
	if ( irqsrc == 0 ) break; // PSW quit when we are done
	if ( (irqsrc & (1<<irq_obj)) == 0 ) continue; // PSW ignore it

	// First reset the IRQ 
        // reread the buffer then reset the irq
	irqsrc2 = CANinw(board, iflag); // PSW reread the iflag
	CANoutw(board, iflag,(1 << irq_obj) );



        //test = (unsigned short)~(1 << irq_obj); 
	//CANsetw(board, iflag, (0xff & ~(1 << irq_obj)));
	//CANoutw(board, iflag, test);
	//irqsrc2 = CANinw(board, iflag);
	//printk(" PSW %d IRQ %d mysource %x source %x obj %d  test %x\n",
	//      tcount,irq,irqsrc,irqsrc2,irq_obj,test);

	
	/*========== receive interrupt */
	
	// PSW Use the Can_BUF to select designated rx buffers
	// get rx_obj from the irq number 
	
	
	if((Can_Buf[irq_obj].status & CAN_BUF_RX)) {
	  ctrl = Can_status_nolock(board,irq_obj);
	  DBGprint(DBG_DATA, (" => got  RX (%d) IRQ[%d]: 0x%0x OID %x\n", 
			      irq, board, irqsrc,CAN_READ_OID(irq_obj)));
	  {
	    //
	    // PSW made this non infinite
	    int rxcount = 0;
	    
	    while ((rxcount < 20 ) && 
		   ((ctrl = Can_status_nolock(board,irq_obj) ) 
		    & (REC_CODE_BUSY << 4)) == (REC_CODE_BUSY << 4)) {
	      /* 20 cycles maximum wait */
	      if (rxcount == 0 ) 
		printk("CAN_int, rx REC_CODE_BUSY"); 
	      rxcount++;
	      udelay(5); // PSW Arbitrary delay
	    }
	  }


	  // PSW we always get our own messages
#if 0
	  //
	  // PSW we let the system recieve its own messages
	  //
	  last_tx_id = CAN_READ_OID_AND_RTR(TRANSMIT_OBJ);
	  if (0 && CAN_READ_OID_AND_RTR(irq_obj) == last_tx_id) {
	    //
	    // PSW what about rtr messages ??
	    /*
	      This is obviously a message we sent ourselves recently.
	      There is no point in delivering this to the upper layer.
	      We rely on the fact
	      that no other node will ever send a message with the same OID.
	    */
	    DBGprint(DBG_DATA, (" PSW own message ignored \n")); 
	    
	  } else {


	    /*---------- kick the select() call  -*/
	    /* __wake_up(struct wait_queue ** p, unsigned int mode); */
	    /* __wake_up(struct wait_queue_head_t *q, unsigned int mode); */
	    /*
	      void wake_up(struct wait_queue**condition)                                                                  
	    */
	    /* This function will wake up all processes
	       that are waiting on this event queue,
	       that are in interruptible sleep
	    */
	    // PSW move this to per device
	    // wake up all devices that are looking at the buffer
	    //wake_up_interruptible(&CanWait[board]); 
	    //
            // PSW add the kill async
            //
	    dev = (can_data_t * )Can_Buf[irq_obj].dev;

	    if (dev)
	      kill_fasync(&dev->fasyncptr, SIGIO, POLL_IN);

	    /* check for CAN controller overrun */
	    /* ToDo */
	  }

#endif

	  
	  //
	  // PSW send to all interested reader processes
	  // 
	  Can_Set_RxFifos (board, irq_obj);
	  
	  //
	  // PSW add the kill async for the buffer
	  //
	  dev = (can_data_t * )Can_Buf[irq_obj].dev;
	  
	  if (dev)
	    kill_fasync(&dev->fasyncptr, SIGIO, POLL_IN);
	  
	  
	  CAN_WRITE_CTRL(irq_obj, REC_CODE_EMPTY, 8);
	  // serviced this irq
	  irqsrc &= ~(1 << irq_obj);
	  DBGprint(DBG_DATA, (" irqsrc now = 0x%0x\n", irqsrc));
	}

	/*========== transmit interrupt */
	// PSW allow any transmit OBJ to be used.
	// get tx_obj from the irq number 
        // PSW use the tx fifo for this buffer	
	TxFifo = &Tx_Bufx[irq_obj];

	if(Can_Buf[irq_obj].status & CAN_BUF_TX) {

	  if (0) 
          {
	    // PSW monitor a bunch of stuff for a while
	    int idtx;
	    //int mymcr[40];
	    for (idtx = 0 ; idtx < 1; idtx++ ) {
	      printk(" MCR(%2d)  0x%04x,", idtx, CANinw(board, canmcr));
	      printk(" ESTAT 0x%04x,", CANinw(board, estat));
	      printk(" IFLAGS 0x%04x \n", CANinw(board, iflag));
	      udelay(10);
	    }
          }

	  ctrl = Can_status_nolock(board,irq_obj);

	  DBGprint(DBG_DATA, (" => got TX (%d) Buf %d IRQ[%d]: 0x%0x OID %x\n",
			      irq, irq_obj, board, 
			      irqsrc,CAN_READ_OID(irq_obj)));

	  // PSW if this was an rtr request then ctrl was set to
          // receive. This totally disabled
          // the buffer for TX. We assume that the user
          // has an RX buffer all ready set up
          // and revert this buffer for Tx.
          // If this is not the case that we will have to
          // think again
          // So we'll simply reset the buffer to TX empty

	  //CAN_WRITE_CTRL(irq_obj, REC_CODE_EMPTY, 8);

	  if ((ctrl & (0xF0)) == (REC_CODE_EMPTY << 4)) {

	    //ctrl = Can_status_lock(board,irq_obj);

            // PSW this will unhang the buffer but only if
	    // it is not needed for rtr rx
	    // the CAN_BUF_RTR_TX status is set in
	    // Can_Send_Common if the MSG_RTR_TX bit
	    // is set in the flags
	    // not that if you have any other 
	    // rxbuffers available they
	    // may receive this message and hang the txbuffer
	    //
	    if((Can_Buf[irq_obj].status & CAN_BUF_RTR_TX)==0) {
	      CAN_WRITE_CTRL(irq_obj, TRANS_CODE_NOT_READY, 8);
	    }
	  } else {
	    ctrl = Can_status_nolock(board,irq_obj);
	  }

	  // PSW add the kill async for the process
	  //
	  dev = (can_data_t * )Can_Buf[irq_obj].dev;
	  
	  if (dev )
	    kill_fasync(&dev->fasyncptr, SIGIO, POLL_OUT);
	  

      	  // PSW moved this here !! 
	  /* Reset Interrupt pending at Transmit Object */
	  //CANsetw(board, iflag, (1 << irq_obj));
	  //printk(" PSW TX buff %d reset ctrl %x \n", 
	  //		      irq_obj, ctrl);	  
          // Reset the TX_BUSY PSW what about rtr ??
	  Can_Buf[irq_obj].status &= ~(CAN_BUF_TX_BUSY) ;

	  /* CAN_register_dump(); */
	  // Check for more data to send from the fifo.
	  if( TxFifo->free[TxFifo->tail] == BUF_EMPTY ) {
	    /* printk("TXE\n"); */
	    TxFifo->status = BUF_EMPTY;
	    TxFifo->active = 0;
	    //printk("PSW TX done\n"); 
	    goto Tx_done;
	  } else {
	    //printk("PSW TX cont...\n"); 
	    ;
	  }
	  
	  
	  /* enter critical section */
	  save_flags(flags);cli();

	  // PSW TODO add a finite delay here
	  tx_wait_delay = 200;
	  


	  while ((Can_status_nolock(board,irq_obj ) & (REC_CODE_BUSY << 4)) == 
		 (REC_CODE_BUSY << 4)) {
	    /* printk("CAN_int, tx REC_CODE_BUSY"); */
	    if ( tx_wait_delay == 0 ) {
	      printk("PSW Error CAN_int, tx REC_CODE_BUSY timeout");
	      
	      break;
	    } 
	    udelay(10);
	    tx_wait_delay--;
	    
	  }
	  /* now we can check here if an RTR was received with this TX chan */
          // PSW we must assume that the MSG_RTR_TX flag was set
	  // this will cause the Can_Buf[irq_obj].status RTR_TX bit
	  // to be set. We will receive one message and then
	  // revert the buffer to Tx
          // use the lock here
	  // Can_status_lock( board,irq_obj ) ; // PSW lock
	  if ((Can_status_lock(board,irq_obj) & (REC_CODE_FULL << 4)) 
	      == (REC_CODE_FULL << 4))
	    {
	      printk("CAN_int, got RX irq on tx channel,  OID= %d\n",
		     CAN_READ_OID(irq_obj));

	      // we send the message to all the rx channel(s) listening
	      Can_Set_RxFifos (board, irq_obj);

	      // Now unlock the buffer for more TX
	      if(Can_Buf[irq_obj].status & CAN_BUF_RTR_TX) {
		CAN_WRITE_CTRL(irq_obj, TRANS_CODE_NOT_READY, 8);
	      }
	    }

	  CAN_SendFifo(board,irq_obj);

	  Can_status_nolock( board,irq_obj ) ; // PSW unlock
	  /* leave critical section */
	  restore_flags(flags);

	} // End TX irq
      Tx_done:
#if 0
	/*========== error status */
	if( irqsrc & CAN_ERROR_INT ) {
	  int s;
	  printk("CAN[%d]: Tx err!\n", board);
	  TxErr[board]++;
	  
	  /* insert error */
	  s = CANin(board,canstat);
	  if(s & CAN_BUS_STATUS )
	    (RxFifo->data[RxFifo->head]).flags += MSG_BUSOFF; 
	  if(s & CAN_ERROR_STATUS)
	    (RxFifo->data[RxFifo->head]).flags += MSG_PASSIVE; 
	  
	  (RxFifo->data[RxFifo->head]).id = 0xFFFFFFFF;
	  /* (RxFifo->data[RxFifo->head]).length = 0; */
	  /* (RxFifo->data[RxFifo->head]).data[i] = 0; */
	  RxFifo->status = BUF_OK;
	  RxFifo->head = ++(RxFifo->head) % MAX_BUFSIZE;
	  if(RxFifo->head == RxFifo->tail) {
	    printk("CAN[%d] RX: FIFO overrun\n", board);
	    RxFifo->status = BUF_OVERRUN;
	  } 
	  
	}
	if( irqsrc & CAN_OVERRUN_INT ) {
	  int s;
	  printk("CAN[%d]: overrun!\n", board);
	  Overrun[board]++;
	  
	  /* insert error */
	  s = CANin(board,canstat);
	  if(s & CAN_DATA_OVERRUN)
	    (RxFifo->data[RxFifo->head]).flags += MSG_OVR; 
	  
	  (RxFifo->data[RxFifo->head]).id = 0xFFFFFFFF;
	  /* (RxFifo->data[RxFifo->head]).length = 0; */
	  /* (RxFifo->data[RxFifo->head]).data[i] = 0; */
	  RxFifo->status = BUF_OK;
	  RxFifo->head = ++(RxFifo->head) % MAX_BUFSIZE;
	  if(RxFifo->head == RxFifo->tail) {
	    printk("CAN[%d] RX: FIFO overrun\n", board);
	    RxFifo->status = BUF_OVERRUN;
	  } 
	  
	  CANout(board, cancmd, CAN_CLEAR_OVERRUN_STATUS );
	} 
#endif

	irqsrc &= ~(1 << irq_obj);
	DBGprint(DBG_DATA, (" irqsrc now = 0x%0x\n", irqsrc));
      
      } // End of all possible irqs
      // PSW no reread    } while( (irqsrc = CANinw(board, iflag)) != 0);
    } while(irqsrc != 0);


/* IRQdone: */
    DBGprint(DBG_DATA, (" => leave IRQ[%d]\n", board));
IRQdone_doneNothing:

}

//
// PSW Handler to send data from the fifo
// needs IRQ s off at the moment
// This will grab the next Tx Message from the fifo and send it.
//  the tx_obj argument is the buffer 
// available to send data to.
// This assumes that there is data in the fifo
//

static void CAN_SendFifo(int board, int tx_obj) 
{
  msg_fifo_t *TxFifo;
  canmsg_t *tx;
  int ret;

  TxFifo = &Tx_Bufx[tx_obj];
  tx = &TxFifo->data[TxFifo->tail];

  /* Real Transmit Interrupt */
  ret = Can_send_common(board, tx, tx_obj);

  TxFifo->free[TxFifo->tail] = BUF_EMPTY; /* now this entry is EMPTY */
  TxFifo->tail = ++(TxFifo->tail) % MAX_BUFSIZE;
  TxFifo->count--; //PSW

}

static int Can_send_common( int board , canmsg_t *tx , int tx_obj )
{

  //tx_obj = tx->buffer;
  if ( tx_obj < 0 || tx_obj > 15 ) {
    printk(" PSW invalid txbuffer used %d correcting to buff 1\n",tx_obj);
    tx_obj = 1;
  }

  tx->length %= 9;			/* limit CAN message length to 8 */
  // Set the busy status 
  Can_Buf[tx_obj].status = CAN_BUF_TX | CAN_BUF_TX_BUSY ;

  // PSW 
  //If the RTR_TX flag is set this means that we want to use this buffer
  // as an rx channel for an rtr message
  // the irq routine will NOT then automatically switch this buffer to 
  // a Tx buffer after sending the RTR message.
  // The buffer will be switched back to Tx after the rtr message has
  // been received
  // Note this will hang this Tx channel if no message is received
  //
  if( tx->flags & MSG_RTR_TX ) {
    Can_Buf[tx_obj].status |= CAN_BUF_RTR_TX;
  }
  /* Writing Control/Status word to hold TX Message Object inactive */
  CAN_WRITE_CTRL(tx_obj, TRANS_CODE_NOT_READY, 1);
  
  /* fill the frame info and identifier fields , ID-Low and ID-High */
  if( tx->flags & MSG_EXT ) {
    /* use ID in extended message format */
    DBGprint(DBG_DATA, ("---> send ext message \n"));
    if( tx->flags & MSG_RTR) {
      CAN_WRITE_XOID_RTR(tx_obj, tx->id);
    } else {
      CAN_WRITE_XOID(tx_obj, tx->id);
    }
  } else {
    DBGprint(DBG_DATA, ("---> send std message \n"));
    if( tx->flags & MSG_RTR) {
      CAN_WRITE_OID_RTR(tx_obj, tx->id);
    } else {
      CAN_WRITE_OID(tx_obj, tx->id);
    }
  }

  memcpy( (char * )&(CAN_OBJ[tx_obj].MSG0),
	  (char * )&tx->data[0],
	  tx->length);
  
  //
  // PSW TxFifo.flags will dictate the operation here
  //
  
  if( tx->flags & MSG_RTR_REP_ONCE) {
    // this one will set the buffer up to send data once for rtr 0xe
    //printk(" PSW set up rtr rep (2) code %x\n",
    //	   TRANS_CODE_TRANSMIT_ONCE_RTR_ALWAYS);
    CAN_WRITE_CTRL(tx_obj, TRANS_CODE_TRANSMIT_ONCE_RTR_ALWAYS,
		   tx->length);

  }else if( tx->flags & MSG_RTR_REP) {
    // this one will set the buffer up to always send this data 0xA
    //printk(" PSW set up rtr rep (3) code %x\n",TRANS_CODE_TRANSMIT_ONLY_RTR);
    CAN_WRITE_CTRL(tx_obj, TRANS_CODE_TRANSMIT_ONLY_RTR,
		   tx->length);
  } else {
    /* Writing Control/Status word (active code, length) */
    CAN_WRITE_CTRL(tx_obj, TRANS_CODE_TRANSMIT_ONCE,
		   tx->length);
  }
  
  return tx->length;
}


void CAN_register_dump(void)
{
volatile flex_can_t *tou_can = (flex_can_t *)(MCF_MBAR + 0x1c0000);

    printk("Flex CAN register layout\n");

#define  printregister(s, name) printk(s, &name , name)
    
    /* printk(" %p: 0x%x \n", tou_can, *(unsigned char *)tou_can); */
    /* printk(" %p: 0x%x \n", (unsigned char *)tou_can + 1, *(((unsigned char *)tou_can) + 1)); */

    printregister(" CAN_ModulConfigRegister      %p %0x\n", CAN_ModulConfigRegister);
    printregister(" CAN_ControlReg0              %p %0x\n", CAN_ControlReg0);
    printregister(" CAN_ControlReg1              %p %0x\n", CAN_ControlReg1);
    printregister(" CAN_ControlReg2              %p %0x\n", CAN_ControlReg2);

    printregister(" CAN_PrescalerDividerRegister %p %0x\n", CAN_PrescalerDividerRegister);
    printregister(" CAN_TimerRegister            %p %0x\n", CAN_TimerRegister);
#if 0
    printregister(" CAN_ReceiveGlobalMask        %p %0x\n", CAN_ReceiveGlobalMask);
    printregister(" CAN_ReceiveBuffer14Mask      %p %0x\n", CAN_ReceiveBuffer14Mask);
    printregister(" CAN_ReceiveBuffer15Mask      %p %0x\n", CAN_ReceiveBuffer15Mask);
#else
    printregister(" CAN_ReceiveGlobalMaskHigh    %p %0x\n", CAN_ReceiveGlobalMaskHigh);
    printregister(" CAN_ReceiveGlobalMaskLow     %p %0x\n", CAN_ReceiveGlobalMaskLow);
    printregister(" CAN_ReceiveBuffer14MaskHigh  %p %0x\n", CAN_ReceiveBuffer14MaskHigh);
    printregister(" CAN_ReceiveBuffer14MaskLow   %p %0x\n", CAN_ReceiveBuffer14MaskLow);
    printregister(" CAN_ReceiveBuffer15MaskHigh  %p %0x\n", CAN_ReceiveBuffer15MaskHigh);
    printregister(" CAN_ReceiveBuffer15MaskLow   %p %0x\n", CAN_ReceiveBuffer15MaskLow);
#endif
    printregister(" CAN_ErrorStatusRegister      %p %0x\n", CAN_ErrorStatusRegister);
    printregister(" CAN_InterruptMasks           %p %0x\n", CAN_InterruptMasks);
    printregister(" CAN_InterruptFlags           %p %0x\n", CAN_InterruptFlags);
    printregister(" CAN_ReceiveErrorCounter      %p %0x\n", CAN_ReceiveErrorCounter);
    printregister(" CAN_TransmitErrorCounter     %p %0x\n", CAN_TransmitErrorCounter);
}


void CAN_object_dump(int object)
{
int board = 0;   /* be prepared for an board paramter, if later on .. */
unsigned int vh;
unsigned int vl;

    volatile unsigned char *cp = (unsigned char *)(MCF_MBAR + 0x1c0080 + (0x10 * object));
    for(vl = 0; vl < 16; vl++) {
	printk("%2x ", *(cp + vl));
    }
    printk("\n");


    printk("Flex CAN object %d\n", object);
    vl = CAN_OBJ[object].ctl_status;
    printk(" Ctrl/Status 0x%x\n", vl);
    vh = CAN_OBJ[object].id_high.id_high;
    printk(" Id high 0x%x\n", vh);
    vl = CAN_OBJ[object].id_low.id_low;
    printk(" Id low 0x%x\n", vl);
    printk(" StdId  %d/0x%x\n", vh >> 5, vh >> 5);
    vh =  (vl >> 1) + ((vh & 7) << 15) + ((vh & 0xffe0) << 13);
    printk(" ExtId  %d/0x%x\n", vh, vh);
}
