/*
 * can_open - can4linux CAN driver module
 *
 * can4linux -- LINUX CAN device driver source
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * 
 * Copyright (c) 2001 port GmbH Halle/Saale
 * (c) 2001 Heinz-Jrgen Oertel (oe@port.de)
 *          Claus Schroeter (clausi@chemie.fu-berlin.de)
 *------------------------------------------------------------------
 * $Header: $
 *
 *--------------------------------------------------------------------------
 *
 *
 * modification history
 * --------------------
 * $Log: can_open.c,v $
 *
 *
 *--------------------------------------------------------------------------
 */


/**
* \file can_open.c
* \author Heinz-Jrgen Oertel, port GmbH
* $Revision: 1.5 $
* $Date: 2002/08/08 18:03:41 $
*
*
*/


/* header of standard C - libraries */
/* #include <linux/module.h>			 */

/* header of common types */

/* shared common header */

/* header of project specific types */

/* project headers */
#include "can_defs.h"

/* local header */

/* constant definitions
---------------------------------------------------------------------------*/

/* local defined data types
---------------------------------------------------------------------------*/

/* list of external used functions, if not in headers
---------------------------------------------------------------------------*/

/* list of global defined functions
---------------------------------------------------------------------------*/

/* list of local defined functions
---------------------------------------------------------------------------*/

/* external variables
---------------------------------------------------------------------------*/

/* global variables
---------------------------------------------------------------------------*/
int Can_isopen[4] = { 0, 0, 0, 0};   /* device minor already opened */

/* local defined variables
---------------------------------------------------------------------------*/
/* static char _rcsid[] = "$Id: can_open.c,v 1.5 2002/08/08 18:03:41 oe Exp $"; */


/***************************************************************************/
/**
*
* \brief int open(const char *pathname, int flags);
* opens the CAN device for following operations
* \param pathname device pathname, usual /dev/can?
* \param flags is one of \c O_RDONLY, \c O_WRONLY or \c O_RDWR which request
*       opening  the  file  read-only,  write-only  or read/write,
*       respectively.
*
*
* The open call is used to "open" the device.
* Doing a first initialization according the to values in the /proc/sys/Can
* file system.
* Additional an ISR function is assigned to the IRQ.
*
* The CLK OUT pin is configured for creating the same frequency
* like the chips input frequency fclk (XTAL). 
*
* If Vendor Option \a VendOpt is set to 's' the driver performs
* an hardware reset befor initializing the chip.
*
* \returns
* open return the new file descriptor,
* or -1 if an error occurred (in which case, errno is set appropriately).
*
* \par ERRORS
* the following errors can occur
* \arg \c ENXIO  the file is a device special file
* and no corresponding device exists.
* \arg \c EINVAL illegal \b minor device number
* \arg \c EINVAL wrong IO-model format in /proc/sys/Can/IOmodel
* \arg \c EBUSY  IRQ for hardware is not available
* \arg \c EBUSY  I/O region for hardware is not available

*/
// get a data buffer index
int get_can_data_idx(void) 
{
  int i;
  for (i= 0 ; i < MAX_CAN_DATA_COUNT ; i ++ ){
    if ( Can_Data[i].id == -1 ) {
      return i;
    }
  }
  return -1;
}

int can_open( __LDDK_OPEN_PARAM )
{
  int retval = 0;
  int can_idx;
    DBGin("can_open");
    {

	int lasterr;	

	unsigned int minor = __LDDK_MINOR;

	if( minor > MAX_CHANNELS )
	{
	    printk("CAN: Illegal minor number %d\n", minor);
	    DBGout();
	    return -EINVAL;
	}
	/* check if device is already open, should be used only by one process */

#ifdef UNSAFE
	/* if(Can_isopen[minor]) { */
	    /* MOD_DEC_USE_COUNT; */
	    /* DBGout(); */
	    /* return -ENXIO; */
	/* } */
	++Can_isopen[minor];		/* flag device in use */
#else

	//	if(Can_isopen[minor] == 1) {
	//    DBGout();
	//    return -ENXIO;
	//}

	
	// PSW check for max number
	if(Can_isopen[minor] >= MAX_CAN_DATA_COUNT-1) {
	  DBGout();
	  return -ENXIO;
	}
	Can_isopen[minor]++;		/* flag device in use */
	//Can_isopen[minor] = 1;		/* flag device in use */
#endif
	if( Base[minor] == 0x00) {
	    /* No device available */
	    printk("CAN[%d]: no device available\n", minor);
	    DBGout();
	    return -ENXIO;
	}



	if ( Can_isopen[minor] == 1 ) { // PSW only do this for first device
	  if( (lasterr = CAN_VendorInit(minor)) < 0 ){
	    DBGout();
	    return lasterr;
	  }
	  /* Access macros based in can_base[] should work now */
	  /* CAN_ShowStat(minor); */

	  /* controller_available(curr + 0x400, 4); */
	  // PSW reset error state
	  Can_state[minor] = 0;

	  Can_WaitInit(minor);	/* initialize wait queue for select() */
	  Can_FifoInit(minor);
	  Can_DataInit(minor);
	  Can_BufInit(minor);
#if CAN_USE_FILTER
	  Can_FilterInit(minor);
#endif

	  if( CAN_ChipReset(minor) < 0 ) {
	    DBGout();
	    return -EINVAL;
	  }
	  CAN_StartChip(minor);
#if DEBUG
    /* CAN_ShowStat(minor); */
    /* CAN_object_dump(0); */
    /* CAN_object_dump(1); */
    /* CAN_object_dump(0); */
    /* CAN_register_dump(); */
#endif
	} // PSW end of System Init

	can_idx = get_can_data_idx();	// PSW check for first open	
	//printk(" PSW can_idx = %d isopen = %d\n",
	//      can_idx,Can_isopen[minor]);       
	/* PSW set up the private data structure for async */
 	file->private_data = &Can_Data[can_idx];

	// PSW set up default read and write buffers
        Can_Data[can_idx].tbuffers = 0x01; //BUF 00 for write
        Can_Data[can_idx].tbuffers = 0x04; //BUF 02 for test write
        Can_Data[can_idx].rbuffers = 0x02; //BUF 01 for read
	Can_Data[can_idx].id = can_idx;  // PSW we loose the id later 
     
	//printk(" PSW tbuf (%p) = %d \n",
	//       &Can_Data[can_idx].tbuffers,
	//       Can_Data[can_idx].tbuffers);
	//printk(" PSW rbuf (%p) = %d \n",
	//       &Can_Data[can_idx].rbuffers,
	//       Can_Data[can_idx].rbuffers);
	
	/* the following does all the board specific things
	   also memory remapping if necessary */

    }

    MOD_INC_USE_COUNT;
    DBGout();
    return retval;
}

