/*
 * can_write - can4linux CAN driver module
 *
 * can4linux -- LINUX CAN device driver source
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * 
 * Copyright (c) 2001 port GmbH Halle/Saale
 * (c) 2001 Heinz-Jrgen Oertel (oe@port.de)
 *          Claus Schroeter (clausi@chemie.fu-berlin.de)
 *------------------------------------------------------------------
 * $Header: $
 *
 *--------------------------------------------------------------------------
 *
 *
 * modification history
 * --------------------
 * $Log: can_write.c,v $
 *
 *
 *
 *
 *--------------------------------------------------------------------------
 */


/**
* \file can_write.c
* \author Heinz-Jrgen Oertel, port GmbH
* $Revision: 1.6 $
* $Date: 2002/08/20 05:57:22 $
*
*/

#include "can_defs.h"

/* \fn size_t can_write( __LDDK_WRITE_PARAM) */
/***************************************************************************/
/**

\brief size_t write(int fd, const char *buf, size_t count);
write CAN messages to the network
\param fd The descriptor to write to.
\param buf The data buffer to write (array of CAN canmsg_t).
\param count The number of bytes to write.

write  writes  up to \a count CAN messages to the CAN controller
referenced by the file descriptor fd from the buffer
starting at buf.



\par Errors

the following errors can occur

\li \c EBADF  fd is not a valid file descriptor or is not open
              for writing.

\li \c EINVAL fd is attached to an object which is unsuitable for
              writing.

\li \c EFAULT buf is outside your accessible address space.

\li \c EINTR  The call was interrupted by a signal before any
              data was written.



\returns
On success, the number of CAN messages written are returned
(zero indicates nothing was written).
On error, -1 is returned, and errno is set appropriately.

\internal
*/
//PSW add to a Fifo
static int can_add_fifo ( int minor, canmsg_t * txp, void * dev)
{
  msg_fifo_t *Fifo;
  int ret;

  ret = 0;
  if ( (txp->buffer < 0) || (txp->buffer >= 16) ) {
    printk (" PSW error in buffer number %d correcting to 0\n",txp->buffer);
    txp->buffer = 0;
  }

  Fifo = &Tx_Bufx[txp->buffer];

  if(Fifo->free[Fifo->head] == BUF_FULL) {
    printk (" PSW Fifo %d full unable to add more \n",txp->buffer);
    return ret;
  }
  ret = 1;
  memcpy(	/* copy one message to FIFO */
	 (char *) &(Fifo->data[Fifo->head]), 
	 (char *) txp,
	 sizeof(canmsg_t) );
  
  //printk(" PSW fifo active: id %d/%x; head %d/tail %d\n",
  //	 (int)Fifo->data[Fifo->head].id,
  //	 (int)Fifo->data[Fifo->head].id, 
  //	 (int)Fifo->head, 
  //	 Fifo->tail); 
  // PSW record who wanted to send the message
  Fifo->dev[Fifo->head] = dev;
  Fifo->free[Fifo->head] = BUF_FULL; /* now this entry is FULL */
  Fifo->head = ++(Fifo->head) % MAX_BUFSIZE;
  
  Fifo->count++;  // this is all that needs crit
  Fifo->active = 1;
  
  return ret;

}

__LDDK_WRITE_TYPE can_write( __LDDK_WRITE_PARAM )
{

unsigned int minor = __LDDK_MINOR;
//msg_fifo_t *TxFifo = &Tx_Buf[minor];
canmsg_t *addr;
canmsg_t tx;
unsigned long flags;
void * dev=file->private_data;    // PSW who sends the message
int written        = 0;
 int ret;                         // PSW ret from send message
    DBGin("can_write");
#ifdef DEBUG_COUNTER
    Cnt1[minor] = Cnt1[minor] + 1;
#endif /* DEBUG_COUNTER */

    if ( Can_state[minor] != 0 ) {
      printk(" PSW invalid Can state %d \n", Can_state[minor]);
      DBGout();
      return -ENODEV;
    }
/* DEBUG_TTY(1, "write: %d", count); */
    //DBGprint(DBG_DATA,(" -- write  %d msg fifo Active %d\n", 
    //		       count,TxFifo->active));
    /* printk("w[%d/%d]", minor, TxFifo->active); */
    addr = (canmsg_t *)buffer;

    if(verify_area(VERIFY_READ, (canmsg_t *)addr, count * sizeof(canmsg_t))) { 
            printk(" PSW verify_area failed \n");
	    DBGout();
	    return -EINVAL;
    }

    while( written < count ) {
	/* enter critical section */
      // PSW why is this critical
	save_flags(flags);cli();
	/* there are data to write to the network */
	//	if(TxFifo->free[TxFifo->head] == BUF_FULL) {
/* DEBUG_TTY(1, "ret buffer full"); */
	    /* there is already one message at this place */;
	//  printk(" PSW Already one message in Tx Fifo\n");
	//  DBGout();
	    /* return -ENOSPC; */
	//  restore_flags(flags); // PSW
	//  return written;
	//}

        // PSW re work this...
	if( 0 /* && TxFifo->active */ ) {
	    /* more than one data and actual data in queue */
	  //__lddk_copy_from_user(	/* copy one message to FIFO */
	  //	    (canmsg_t *) &(TxFifo->data[TxFifo->head]), 
	  //	    (canmsg_t *) &addr[written],
	  //	    sizeof(canmsg_t) );
	    //printk(" PSW fifo active: id %d/%x; head %d/tail %d\n",
	    //	   TxFifo->data[TxFifo->head].id,
	    //	   TxFifo->data[TxFifo->head].id, 
	    //	   TxFifo->head, 
	    //	   TxFifo->tail); 
	    // PSW record who wanted to send the message
	  // TxFifo->dev[TxFifo->head] = dev;
	  // TxFifo->free[TxFifo->head] = BUF_FULL; /* now this entryisFULL */
	  //   TxFifo->head = ++(TxFifo->head) % MAX_BUFSIZE;
    	  //  TxFifo->count++;  // this is all that needs crit
	} else {

	    __lddk_copy_from_user(
		    (canmsg_t *) &tx, 
		    (canmsg_t *) &addr[written],
		    sizeof(canmsg_t) );
/* DEBUG_TTY(1, " fifo in-active: id %d/%x; head %d/tail %d", */
		/* tx.id, tx.id, */
/* TxFifo->head, */
/* TxFifo->tail ); */
	  /* f - fast -- use interrupts */
	    //	  if(0 && ( count >= 1) ) {
	    ///* !!! CHIP abh. !!! */
	    ////printk(" Setting TxFifo->active (was %d) written = %d\n",
	    //   TxFifo->active,count);
	    //TxFifo->active = 1;
	    //}
	    //PSW test for an RX buffer we cannot write to them
	    if (Can_Buf[tx.buffer].status & CAN_BUF_RX) {
	      printk(" PSW cannot write to RX buffer %d \n",tx.buffer);
	      printk(" PSW skipping write request\n");

	      ret = 1 ;
	    } else {
	      ret = CAN_SendMessageX( minor, &tx, dev);  /* Send, no wait */
	    }
	  //printk(" PSW called CAN_SendMessageX ret = %d\n",ret);

	  // PSW TODO If this fails then add to Fifo
          if ( ret == 0 ) 
	    ret = can_add_fifo (minor, &tx, dev);
	  
	}
        written++;
	/* leave critical section */
	restore_flags(flags);
    }
    /* printk("W[%d/%d]", minor, TxFifo->active); */
    DBGout();
    return written;
}

