/* can driver
*  test programm
*  send messages commissioned via command line
*  this is a simple system to send a basic message every second
*  it uses a fasync handler to detect when the message is sent.
*
*  PSW heavily derived from can_send
* UCLINUX_BUILD_USER=1 make user/can4linux_only  
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
/* #include <sys/time.h> */
#include <time.h>
#include <sys/ioctl.h>
#include <sys/poll.h>


#ifdef USE_RT_SCHEDULING
# include <sched.h>
# include <sys/mman.h>
#endif

#include <errno.h>
#include "can4linux.h"

#define    DEBUG_NONE  ioctl(can_fd, CAN_DEBUG_LEVEL, CAN_DEBUG_NONE)
#define    DEBUG_SOME  ioctl(can_fd, CAN_DEBUG_LEVEL, CAN_DEBUG_SOME)
#define    DEBUG_ALL   ioctl(can_fd, CAN_DEBUG_LEVEL, CAN_DEBUG_ALL)

#define STDDEV "/dev/can0"
#define CANDEV "/dev/can0"
#define VERSION "1.6"

#ifndef TRUE
# define TRUE  1
# define FALSE 0
#endif

extern int errno;
canmsg_t message[4];  // 4 messages
canmsg_t rtrmessage[4];  // 4 messages
canmsg_t rxmessage[4];  // 4 messages

int	set_rbuffer(int fd,int rbuffer);

//
// PSW set up an rtr reply for message 0x250
// we will use buffer 5 for this 1<<5
//
int set_rtr_reply(int fd) 
{
  int ix;
  int rbuffer;
  int ret;

  ix = 0;
  /* our default 8 byte message */
  rtrmessage[ix].id      = 0x255;  // this matters
  rtrmessage[ix].buffer  = 8;   /* txbuffer  */
  rtrmessage[ix].cob     = 0;
  rtrmessage[ix].flags   = MSG_RTR_REP; // this is a rtr reply
  //rtrmessage[ix].flags   = 0; // this is not an rtr reply
  rtrmessage[ix].length  = 8;
  rtrmessage[ix].data[0] = 0xee;
  rtrmessage[ix].data[1] = 0x11;
  rtrmessage[ix].data[2] = 0xee;
  rtrmessage[ix].data[3] = 0x00;
  rtrmessage[ix].data[4] = 0xee;
  rtrmessage[ix].data[5] = 0x11;
  rtrmessage[ix].data[6] = 0Xee;
  rtrmessage[ix].data[7] = 0x11;
  

  rbuffer = (1<<5); // PSW set a rx buffer mask to get tx  
  set_rbuffer(fd,rbuffer);
  ret = write(fd, &rtrmessage[0], 1);

  // or use ioctl call
  //ret = can_send(fd, &rtrmessage[0]);
  //printf("PSW send rtr 1 ret = %d \n",ret);

  // test overwrite
  //rtrmessage[ix].data[1] = 0x88;
  //ret = write(fd, &rtrmessage[0], 1);
  //ret = can_send(fd, &rtrmessage[0]);
  //printf("PSW send rtr 2 ret = %d \n",ret);



  return ret;

}

//
// Example simple async handler
//  
void can_async_handler(int signo)
{
  /* got an alarm, exit & recycle */
  printf(" SIGIO received \n");
}  

//
// set the general mask
//
int	set_acc_mask(
	int fd	,		/* device descriptor */
	int mask
	)
{
Config_par_t  cfg;

    cfg.target = CONF_ACC; 
    cfg.val1   = mask;
    ioctl(fd, CONFIG, &cfg);
    return 0;

}
//
// set buff OID
//
int	set_buff_oid(
	int fd	,		/* device descriptor */
	int buff,
	int oid
	)
{
Config_par_t  cfg;

    cfg.target = CONF_BUFFMASK; 
    cfg.val1   = buff;
    cfg.val2   = oid;
    ioctl(fd, CONFIG, &cfg);
    return 0;
}
//
// set buff extended OID
//
int	set_buff_eoid(
	int fd	,		/* device descriptor */
	int buff,
	int oid
	)
{
Config_par_t  cfg;

    cfg.target = CONF_BUFFMASKE; 
    cfg.val1   = buff;
    cfg.val1   = oid;
    ioctl(fd, CONFIG, &cfg);
    return 0;
}

//
//  set both rbuffers and tbuffers
//  this  are masks of transmit buffers and recieve buffers
//    we can use for this device instance
//
int	set_rbuffer(
	int fd,			/* device descriptor */
	int rbuffer)            /* masks   */

{
Config_par_t  cfg;


    cfg.target = CONF_SETSELBUFS; 
    cfg.val1=rbuffer;
    cfg.val2=0;
    ioctl(fd, CONFIG, &cfg);

    return 0;
}

//
// Set the baud rate
//
int	set_bitrate(
	int fd,			/* device descriptor */
	int baud		/* bit rate */
	)
{
  Config_par_t  cfg;
  volatile Command_par_t cmd;
  int ret;

  cmd.cmd = CMD_STOP;
  ioctl(fd, COMMAND, &cmd);
  
  cfg.target = CONF_TIMING; 
  cfg.val1   = baud;
  ioctl(fd, CONFIG, &cfg);
  
  cmd.cmd = CMD_START;
  ret=ioctl(fd, COMMAND, &cmd);
  return ret;
}

//
// Send default message
//


int send_default_message ( int fd , int num ) 
{
  int ret;
  int ix = 0;

  if ( num  > 4 ) num = 4;

  message[ix].id      = 0x101;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  message[ix].flags   = 0; // this is not an rtr reply
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x10;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x102;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  message[ix].flags   = 0; // this is not an rtr reply
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x20;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x103;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  message[ix].flags   = 0; // this is not an rtr reply
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x30;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x104;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  message[ix].flags   = 0; // this is not an rtr reply
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x40;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;

  ret = write(fd,message,num);

  return ret;
}//
// Send default message
//


int send_default_message_rtr ( int fd , int num ) 
{
  int ret;
  int ix = 0;

  if ( num  > 4 ) num = 4;

  message[ix].id      = 0x255;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  message[ix].flags   = MSG_RTR; // this is an rtr  request
  //message[ix].flags   = 0; // this is not an rtr reply
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x10;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x256;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  //message[ix].flags   = 0; // this is not an rtr reply
  message[ix].flags   = MSG_RTR; // this is an rtr  request
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x20;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x257;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  //message[ix].flags   = 0; // this is not an rtr reply
  message[ix].flags   = MSG_RTR; // this is an rtr  request
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x30;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;
  ix++;
  message[ix].id      = 0x258;
  message[ix].buffer  = 7;   /* txbuffer  */
  message[ix].cob     = 0;
  //message[ix].flags   = 0; // this is not an rtr reply
  message[ix].flags   = MSG_RTR; // this is an rtr  request
  message[ix].length  = 8;
  message[ix].data[0] = 0x10;
  message[ix].data[1] = 0x40;
  message[ix].data[2] = 0x22;
  message[ix].data[3] = 0x33;
  message[ix].data[4] = 0x44;
  message[ix].data[5] = 0x55;
  message[ix].data[6] = 0x66;
  message[ix].data[7] = 0x77;

  ret = write(fd,message,num);

  return ret;
}
//
//
//
int print_rx_message ( int ix, int fd ) 
{
  int ret;
  ret = 0;
  printf(
     " RX (%d) fd %d [%04x] %d %02x %02x %02x %02x %02x %02x %02x %02x \n",
     rxmessage[ix].buffer,
     fd,
     (int)rxmessage[ix].id,
     rxmessage[ix].length,
     rxmessage[ix].data[0],
     rxmessage[ix].data[1],
     rxmessage[ix].data[2],
     rxmessage[ix].data[3],
     rxmessage[ix].data[4],
     rxmessage[ix].data[5],
     rxmessage[ix].data[6],
     rxmessage[ix].data[7]);
  
  
  return ret;
}

//
// reset the device
//
int can_reset( int fd ) 
{
  int ret;
  volatile Command_par_t cmd;

  cmd.cmd = CMD_RESET;
  ret = ioctl(fd, COMMAND, &cmd);
  
  return ret;
}


int test1(void )
{

  int fd;
  int ret;
  int baud;

  printf("*****************************************************\n");
  printf(" Starting Test 1 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");

    exit (1);
  }

  //
  // Set the speed normally only do this once 
  // try 500 first
  //
  baud = 500;
  set_bitrate(fd,baud);
  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);

  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);

  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  //
  // close the device 
  //
  close(fd);

  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");

    exit (1);
  }

  //
  // Set the speed normally only do this once 
  // try 125 first
  //
  set_bitrate(fd,125);
  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);
  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);
  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  close(fd);

  //////////////////////////////////////////////////
  // Just change bit rate
  //////////////////////////////////////////////////
  
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    
    exit (1);
  }
  
  //
  // Set the speed normally only do this once 
  // try 500 first
  //
  baud = 500;
  set_bitrate(fd,baud);
  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);
  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);
  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  baud = 125;
  set_bitrate(fd,baud);

  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);
  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);
  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  close(fd);

  //////////////////////////////////////////////////
  // change bit rate and reset chip
  //////////////////////////////////////////////////
  
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");

    exit (1);
  }

  //
  // Set the speed normally only do this once 
  // try 500 first
  //
  baud = 500;
  set_bitrate(fd,baud);
  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);
  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);
  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  baud = 125;
  set_bitrate(fd,baud);
  can_reset(fd);
  
  //
  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);
  printf(" Message 1 sent at Rate %d ret = %d \n",baud,ret);
  ret = send_default_message(fd,1);
  printf(" Message 2 sent at Rate %d ret = %d \n",baud,ret);

  close(fd);


  return 0;
}

/////////////////////////////////////////////////////
// test 2 async and poll set up for tx
// open with no baud rate change for this
////////////////////////////////////////////////////
int test2(void )
{
  struct sigaction act;

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  int fdflags;
  int i;

  printf("*****************************************************\n");
  printf(" Starting Test 2 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }

  //
  // Set the speed normally only do this once 
  // try 500 first
  //
  //baud = 500;
  //set_bitrate(fd,baud);
  //

  // set up Async Handler
  
  act.sa_handler = can_async_handler;
  sigemptyset(&(act.sa_mask));
  act.sa_flags = 0;

  
  sigaction(SIGIO, &act, NULL);
  
  fcntl(fd, F_SETOWN, getpid());
  fdflags = fcntl(fd, F_GETFL);
  fcntl(fd, F_SETFL, fdflags | FASYNC );

  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLOUT ;
  num_fds = 1;
  timeout = 150; // 150 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  ret = poll(pollfd, 1, timeout);
  switch ( ret) {
  case -1:
    printf(" PSW Poll error \n");
    break;

  case 0:
    printf(" PSW Poll timeout Correct\n");
    break;

  default:
    printf(" PSW Poll data %d \n",ret);
    for ( i = 0 ; i < num_fds ; i++ ) {
      
      if (pollfd[i].revents == POLLIN ) {
	printf(" data ready on %d (fd %d)\n",
	       i,pollfd[i].fd );
      } else if (pollfd[i].revents == POLLOUT ) {
	printf(" data ready on %d (fd %d)\n",
	       i,pollfd[i].fd );
      } else {

      }
      
    }    
    break;
  }
  

  // try to send a message to OID 100
  //
  ret = send_default_message(fd,1);

  printf(" Test 2 Message 1 sent at ret = %d \n",ret);
  //
  // close the device 
  //
  close(fd);
  return 0;

}

/////////////////////////////////////////////////////
// test 3 poll rx set 
// open with no baud rate change for this
////////////////////////////////////////////////////
int test3(void )
{

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  int num_msg;
  int total_time;
  int item;
  printf("*****************************************************\n");
  printf(" Starting Test 3 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }
  // Wide open
  set_acc_mask(fd,0x0);

  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLIN ;
  num_fds = 1;
  timeout = 500; // 500 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  num_msg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds

    ret = poll(pollfd, 1, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = read(pollfd[i].fd,&rxmessage[0],1);
	  if ( item > 0 ) {
	    print_rx_message (0,pollfd[i].fd); 
	    num_msg++;
	  } else {
	    printf("Message error \n");
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 3 %d Messages received \n",num_msg);
  //
  // close the device 
  //
  close(fd);

  return num_msg;
}

/////////////////////////////////////////////////////
// test 4 poll rx set
// set mask filter
// rx on OID 0x200 only 
////////////////////////////////////////////////////
int test4(void )
{

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  int num_msg;
  //  int num_rxmsg;
  int total_time;
  int item;
  printf("*****************************************************\n");
  printf(" Starting Test 4 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }

  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLIN ;
  num_fds = 1;
  timeout = 500; // 500 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  set_acc_mask(fd,0xffffffff);
  set_buff_oid(fd,1,0x200);



  num_msg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds

    ret = poll(pollfd, 1, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = read(pollfd[i].fd,&rxmessage[0],1);
	  if ( item > 0 ) {
	    print_rx_message (0,pollfd[i].fd); 
	    num_msg++;
	  } else {
	    printf("Message error \n");
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 4 %d Messages received \n",num_msg);
  //
  // close the device 
  //
  close(fd);

  return num_msg;
}


/////////////////////////////////////////////////////
// test 5 multiple rx buffers 
// set mask filter
// rx 0 on all of them 
// rx 1 on OID 0x200 only 
// rx 2 on OID 0x300 only 
// rx 3 on OID 0x400 only 
////////////////////////////////////////////////////
int test5(void )
{

  int fd[4];
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  int num_msg;
  int total_time;
  int item;
  printf("*****************************************************\n");
  printf(" Starting Test 5 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  for ( i = 0 ; i < 4 ; i++ ) {
    fd[i] = open(CANDEV,O_RDWR);
    
    if ( fd[i] < 0 ) {
      printf(" failed to open device (%s) %d ..\n",CANDEV,i);
      printf(" check the /dev directory \n");
      exit (1);
    }
    pollfd[i].fd = fd[i];
    pollfd[i].events = POLLIN ;
  }
  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  num_fds = 4;
  timeout = 500; // 500 mS

  printf(" PSW Poll num fds %d timeout %d POLLIN %x\n",
	 num_fds,timeout,POLLIN);

  set_acc_mask(fd[0],0xffffffff);
  set_buff_oid(fd[0],1,0x200);  // Note the fd number is not important here 
  set_buff_oid(fd[0],2,0x300);  // Note that this also turns the buffer
  set_buff_oid(fd[0],3,0x400);  // into a rx buffer

  // now set the interest masks for each fd this time the fd does count

  set_rbuffer(fd[0],(1<<1) | (1<<2) | (1<<3)  );
  set_rbuffer(fd[1], (1<<1) );
  set_rbuffer(fd[2], (1<<2) );
  set_rbuffer(fd[3], (1<<3) );


  num_msg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds

    ret = poll(pollfd, num_fds, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      printf(" \n *** PSW Poll ret %d \n",ret);
      for ( i = 0 ; i < num_fds ; i++ ) {
	//	printf(" PSW Poll fd %d revents %x  \n",
	//       pollfd[i].fd,pollfd[i].revents);
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = 1;
          while ( item ) {
	    item = read(pollfd[i].fd,&rxmessage[0],1);
	    if ( item > 0 ) {
	      print_rx_message (0,pollfd[i].fd); 
	      num_msg++;
	    } else {
	      if ( item < 0 )
		printf("Message error \n");
	    }
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 5 %d Messages received \n",num_msg);
  //
  // close the device 
  //
  close(fd[0]);
  close(fd[1]);
  close(fd[2]);
  close(fd[3]);

  return num_msg;
}


/////////////////////////////////////////////////////
// test 6 multiple tx messages
////////////////////////////////////////////////////
int test6(void )
{

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  //  int num_msg;
  int total_time;
  int item;
  int num_txmsg;
  int num_rxmsg;
  printf("*****************************************************\n");
  printf(" Starting Test 6 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }

  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLIN ;
  num_fds = 1;
  timeout = 500; // 500 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  set_acc_mask(fd,0x00000000); // wide open
  set_buff_oid(fd,1,0x200);    // dont care





  num_rxmsg = 0;
  num_txmsg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds
    num_txmsg += send_default_message(fd,4);  
    num_txmsg += send_default_message(fd,4);
    ret = poll(pollfd, 1, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = 1;
          while ( item ) {
	    item = read(pollfd[i].fd,&rxmessage[0],1);
	    if ( item > 0 ) {
	      print_rx_message (0,pollfd[i].fd); 
	      num_rxmsg++;
	    } else {
	      if ( item < 0 )
		printf("Message error \n");
	    }
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 6 %d Messages sent %d received\n",num_txmsg,num_rxmsg);
  //
  // close the device 
  //
  close(fd);

  return num_txmsg;
}
/////////////////////////////////////////////////////
// test 7 multiple tx rtr  messages
// send on ch 251 and just filter these replies
////////////////////////////////////////////////////
int test7(void )
{

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  //  int num_msg;
  int total_time;
  int item;
  int num_txmsg;
  int num_rxmsg;

  printf("*****************************************************\n");
  printf(" Starting Test 7 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }

  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLIN ;
  num_fds = 1;
  timeout = 500; // 500 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  set_acc_mask(fd,0xffffffff); // closed
  set_buff_oid(fd,1,0x255);    // dont care





  num_rxmsg = 0;
  num_txmsg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds
    num_txmsg += send_default_message_rtr(fd,1);  

    ret = poll(pollfd, 1, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = 1;
          while ( item ) {
	    item = read(pollfd[i].fd,&rxmessage[0],1);
	    if ( item > 0 ) {
	      print_rx_message (0,pollfd[i].fd); 
	      num_rxmsg++;
	    } else {
	      if ( item < 0 )
		printf("Message error \n");
	    }
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 7 %d Messages sent %d received\n",num_txmsg,num_rxmsg);
  //
  // close the device 
  //
  close(fd);

  return num_txmsg;
}
/////////////////////////////////////////////////////
// test 8
// no rtr set up 
// wide open listen
////////////////////////////////////////////////////
int test8(void )
{

  int fd;
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  //  int num_msg;
  int total_time;
  int item;
  int num_txmsg;
  int num_rxmsg;

  printf("*****************************************************\n");
  printf(" Starting Test 8 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //
  fd = open(CANDEV,O_RDWR);

  if ( fd < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }

  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd;
  pollfd[0].events = POLLIN ;
  num_fds = 1;
  timeout = 500; // 500 mS
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  set_acc_mask(fd,0x00000000); // open
  set_buff_oid(fd,1,0x255);    // dont care

  num_rxmsg = 0;
  num_txmsg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds
    //num_txmsg += send_default_message_rtr(fd,1);  

    ret = poll(pollfd, 1, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = 1;
          while ( item ) {
	    item = read(pollfd[i].fd,&rxmessage[0],1);
	    if ( item > 0 ) {
	      print_rx_message (0,pollfd[i].fd); 
	      num_rxmsg++;
	    } else {
	      if ( item < 0 )
		printf("Message error \n");
	    }
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
	
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 8 %d Messages sent %d received\n",num_txmsg,num_rxmsg);
  //
  // close the device 
  //
  close(fd);

  return num_txmsg;
}
/////////////////////////////////////////////////////
// test 9
// rtr response
// Set up buffer 8 to reply to an RTR request
////////////////////////////////////////////////////
int test9(void )
{

  int fd[4];
  int ret;

  struct pollfd pollfd[16];
  int num_fds;
  int timeout;
  //int fdflags;
  int i;
  //  int num_msg;
  int total_time;
  int item;
  int num_txmsg;
  int num_rxmsg;

  printf("*****************************************************\n");
  printf(" Starting Test 9 \n");
  printf("*****************************************************\n");

  //
  // Open the device
  //

  fd[0] = open(CANDEV,O_RDWR);

  if ( fd[0] < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }
  fd[1] = open(CANDEV,O_RDWR);

  if ( fd[1] < 0 ) {
    printf(" failed to open device (%s) ..\n",CANDEV);
    printf(" check the /dev directory \n");
    exit (1);
  }


  // set accmask
  // set up a poll
  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  pollfd[0].fd = fd[0];
  pollfd[0].events = POLLIN | POLLOUT;
  pollfd[1].fd = fd[1];
  pollfd[1].events = POLLIN | POLLOUT;
  num_fds = 2;
  timeout = 1000; // 1 S
  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  //set_acc_mask(fd[0],0xffffffff); // closed
  set_acc_mask(fd[0],0x00000000); // open
  set_buff_oid(fd[0],1,0x255);    // dont care


  set_rtr_reply(fd[0]); // set up the rtr buffer 
  //write(fd[0], &rtrmessage[0], 1);

  num_rxmsg = 0;
  num_txmsg = 0;
  total_time = 0;
  while ( total_time < 5000 ) {  // 5 seconds
    //num_txmsg += send_default_message_rtr(fd,1);  
    //num_txmsg += write(fd[0], &rtrmessage[0], 1);

    ret = poll(pollfd, num_fds, timeout);
    switch ( ret) {
    case -1:
      printf(" PSW Poll error \n");
      break;
      
    case 0:
      printf(" PSW Poll timeout ...total time %d \n",total_time);
      break;

    default:
      for ( i = 0 ; i < num_fds ; i++ ) {
	
	if (pollfd[i].revents == POLLIN ) {
	  printf(" data in ready on %d (fd %d)\n",
		 i,pollfd[i].fd );
	  item = 1;
          while ( item ) {
	    item = read(pollfd[i].fd,&rxmessage[0],1);
	    if ( item > 0 ) {
	      print_rx_message (0,pollfd[i].fd); 
	      num_rxmsg++;
	    } else {
	      if ( item < 0 )
		printf("Message error \n");
	    }
	  }
	} else if (pollfd[i].revents == POLLOUT ) {
	  printf(" PSW Poll OUT !!!\n");
	} else {
	  ;
	}
      }    
      break;
    }
    total_time += timeout;
  }
  printf(" Test 9 %d Messages sent %d received\n",num_txmsg,num_rxmsg);
  //
  // close the device 
  //
  close(fd);

  return num_txmsg;
}

//
// The main program
//

int main(int argc, char **argv)
{
#if 1
  test1();
  test2();
  test3();
  test4();
  test5();
  test6();
  test7(); 
  test8(); 
#endif
  test9(); 
 return 0;
}
#if 0


canmsg_t message[4];  // default send messages
canmsg_t rtrmessage[4];  // rtr reply message
canmsg_t rmessage[10]; // default read message
void sbuf(unsigned char *s, int n);
void usage(char *s);
int can_fd;
int can_fd2;
int can_fd3;  // for rtr

int node                 = 8;
int debug                = FALSE;
int extd                 = FALSE;
int load		 = 50;		/* default bus-load */
int doload		 = FALSE;		/* default bus-load */
int stresstest           = FALSE;
int testtype             = 1;  
int sleeptime            = 1000;	/* one second */
int cstdout              = FALSE;	/* use stdout for CAN message */
long int test_count_max = 0;
int rtr                  = FALSE;
int baud		 = 500;		/* set  baud rate */
int priority		 = -1;		/* dont change priority rate */
int canreset		 = FALSE;
int use_async            = 1;           /* use async handler */

/* functions */
void clean_process(void);
void test1(void);

/***********************************************************************
*
* set_bitrate - sets the CAN bit rate
*
*
* Changing these registers only possible in Reset mode.
*
* RETURN:
*
*/

//
// Set the baud rate
//
int	set_bitrate(
	int fd,			/* device descriptor */
	int baud		/* bit rate */
	)
{
Config_par_t  cfg;
volatile Command_par_t cmd;
 int ret;

    cmd.cmd = CMD_STOP;
    ioctl(fd, COMMAND, &cmd);

    cfg.target = CONF_TIMING; 
    cfg.val1   = baud;
    ioctl(fd, CONFIG, &cfg);

    cmd.cmd = CMD_START;
    ret=ioctl(fd, COMMAND, &cmd);
    return ret;
}

int	get_rbuffer(
	int fd			/* device descriptor */
	)
{
Config_par_t  cfg;

    cfg.target = CONF_GETSELBUFS; 
    ioctl(fd, CONFIG, &cfg);

     return cfg.val1;
}


//
// get the tbuffers
//  this is a mask of transmit buffers we can use for this instance
//
int	get_tbuffer(
	int fd			/* device descriptor */
	)
{
Config_par_t  cfg;

    cfg.target = CONF_GETSELBUFS; 
    ioctl(fd, CONFIG, &cfg);

     return cfg.val2;
}

//
//  set both rbuffers and tbuffers
//  this  are masks of transmit buffers and recieve buffers
//    we can use for this instance
//
int	set_rtbuffer(
	int fd,			/* device descriptor */
	int rbuffer,            /* masks   */
	int tbuffer)
{
Config_par_t  cfg;


    cfg.target = CONF_SETSELBUFS; 
    cfg.val1=rbuffer;
    cfg.val2=tbuffer;
    ioctl(fd, CONFIG, &cfg);

    return 0;
}

/*  
  get the device status
*/
void getStat(void)
{
CanSja1000Status_par_t status;
    ioctl(can_fd, STATUS, &status);
    printf(":: %02x %d %d %d %02x | %d | r%d t%d\n",
    	status.status,
    	status.error_warning_limit,
    	status.rx_errors,
    	status.tx_errors,
    	status.error_code,
    	status.rx_buffer_size,
    	status.rx_buffer_used,
    	status.tx_buffer_used
    	);
}

//
// reset the device
//
int can_reset( void ) {

int ret;
volatile Command_par_t cmd;


    cmd.cmd = CMD_RESET;
    ret = ioctl(can_fd, COMMAND, &cmd);

    return ret;
}

//
// send buffer via ioctl
//
int can_send(int fd, canmsg_t * tx) 
{

int ret;
volatile Message_par_t cmd;

    memcpy((char *)&cmd.Tx,(char *)tx,sizeof(canmsg_t));
    ret = ioctl(fd, SEND, &cmd);

    return ret;
}


//
// Example simple async handler
//  
void can_async_handler(int signo)
{
  /* got an alarm, exit & recycle */
  printf(" SIGIO received \n");
}  

//
// Example poll handler
// given a list of fds will wait for the first one to receive
//
// use
//  int fds[16];
//  int ret;
//
//  fds[0] = can_fd;
//  fds[1] = can_fd2;
//  ret = can_poll(fds,2,10);  
//
int can_poll ( int *fds , int num_fds, int timeout)
{
  struct pollfd pollfd[16];
  int ret;
  int i;

  printf(" PSW Poll num fds %d timeout %d\n",num_fds,timeout);

  //
  // fill out the pollfd structure
  // in this test we assume a read
  //

  for ( i = 0 ; i < num_fds ; i++ ) {
    pollfd[i].fd = fds[i];
    pollfd[i].events = POLLIN ;
  }

  ret = poll(pollfd, num_fds, timeout);
  switch ( ret) {
  case -1:
    printf(" PSW Poll error \n");
    break;

  case 0:
    printf(" PSW Poll timeout \n");
    break;

  default:
    printf(" PSW Poll data %d \n",ret);
    for ( i = 0 ; i < num_fds ; i++ ) {

      if (pollfd[i].revents == POLLIN ) {
	printf(" data ready on %d (fd %d)\n",
	          i,pollfd[i].fd );
      } else {
	// Dummy set the fd to 0 if there is not data
	fds[i] = 0;
      }
    }    
    break;
  }

  return ret;

}

//
// PSW reads input buffers
//
int read_buff ( int fd , char * bname)
{
  int numreads = 0;
  int totreads = 0;
  int ret;
  
  while ( (ret = read(fd, &rmessage[0], 10)) > 0 ) {
    numreads++;
    totreads += ret;
    //printf("read_fd (%d) %d\n", numreads,ret);
  }
  printf("read_fd (%s) total %d\n", bname,totreads);
  return totreads;
}


//
// PSW set up an rtr reply for message 0x250
// we will use buffer 5 for this 1<<5
//
int set_rtr_reply(int fd) 
{
  int ix;
  int rbuffer;
  int tbuffer;
  int ret;

  ix = 0;
  /* our default 8 byte message */
  rtrmessage[ix].id      = 0x250;
  rtrmessage[ix].buffer  = 8;   /* txbuffer  */
  rtrmessage[ix].cob     = 0;
  rtrmessage[ix].flags   = MSG_RTR_REP; // this is a rtr reply
  //rtrmessage[ix].flags   = 0; // this is not an rtr reply
  rtrmessage[ix].length  = 8;
  rtrmessage[ix].data[0] = 0x00;
  rtrmessage[ix].data[1] = 0x11;
  rtrmessage[ix].data[2] = 0x22;
  rtrmessage[ix].data[3] = 0x33;
  rtrmessage[ix].data[4] = 0x44;
  rtrmessage[ix].data[5] = 0x55;
  rtrmessage[ix].data[6] = 0x66;
  rtrmessage[ix].data[7] = 0x77;
  
  // PSW get/set the masks

  //rbuffer = get_rbuffer(fd);
  //tbuffer = get_tbuffer(fd);
      
  //printf("PSW rbuffer = %x tbuffer = %x/n",rbuffer,tbuffer);
  tbuffer = (1<<5); // PSW set a tx buffer mask
  rbuffer = (1<<5); // PSW set a rx buffer mask to get tx  
  set_rbuffer(fd,rbuffer,tbuffer);
  // You cannot do this yet with an RTR request
  ret = write(fd, &rtrmessage[0], 1);
  // use ioctl call
  //ret = can_send(fd, &rtrmessage[0]);
  printf("PSW send rtr 1 ret = %d \n",ret);
  rtrmessage[ix].data[1] = 0x88;
  ret = write(fd, &rtrmessage[0], 1);
  //ret = can_send(fd, &rtrmessage[0]);
  printf("PSW send rtr 2 ret = %d \n",ret);



  return 0;

}
/**
*
* The main program
*
*/
int main(int argc, char **argv)
{
int ret;
int cnt;
int c;
char *pname;
extern char *optarg;
extern int optind;
char device[40] = STDDEV;
int max_priority;
int increment = 0;i
struct sigaction act;
int ix;

// for can_poll test
int fds[16];



    pname = *argv;

#ifdef USE_RT_SCHEDULING
    max_priority = sched_get_priority_max(SCHED_RR) - 1;
#else 
    max_priority = 1;
#endif
    ix = 0;
    /* our default 8 byte message */
    message[ix].id      = 200;
    message[ix].buffer  = 6;   /* txbuffer  */
    message[ix].cob     = 0;
    message[ix].flags   = 0;
    message[ix].length  = 8;
    message[ix].data[0] = 0x55;
    message[ix].data[1] = 0;
    message[ix].data[2] = 3;
    message[ix].data[3] = 4;
    message[ix].data[4] = 5;
    message[ix].data[5] = 6;
    message[ix].data[6] = 7;
    message[ix].data[7] = 0xaa;

    ix = 1;
    message[ix].id      = 201;
    message[ix].cob     = 0;
    message[ix].buffer  = 6;   /* txbuffer  */
    message[ix].flags   = 0;
    message[ix].length  = 8;
    message[ix].data[0] = 0xaa;
    message[ix].data[1] = 0;
    message[ix].data[2] = 3;
    message[ix].data[3] = 4;
    message[ix].data[4] = 5;
    message[ix].data[5] = 6;
    message[ix].data[6] = 7;
    message[ix].data[7] = 0x55;
    ix = 2;
    message[ix].id      = 202;
    message[ix].cob     = 0;
    message[ix].buffer  = 6;   /* txbuffer  */
    message[ix].flags   = 0;
    message[ix].length  = 8;
    message[ix].data[0] = 0xaa;
    message[ix].data[1] = 0;
    message[ix].data[2] = 3;
    message[ix].data[3] = 4;
    message[ix].data[4] = 5;
    message[ix].data[5] = 6;
    message[ix].data[6] = 7;
    message[ix].data[7] = 0x55;
    ix = 3;
    message[ix].id      = 203; // PSW
    message[ix].cob     = 0;
    message[ix].buffer  = 6;   /* txbuffer  */
    message[ix].flags   = 0;
    message[ix].length  = 8;
    message[ix].data[0] = 0xaa;
    message[ix].data[1] = 0;
    message[ix].data[2] = 3;
    message[ix].data[3] = 4;
    message[ix].data[4] = 5;
    message[ix].data[5] = 6;
    message[ix].data[6] = 7;
    message[ix].data[7] = 0x55;


    while ((c = getopt(argc, argv, "ab:dehl:rp:s:n:D:t:T:VR")) != EOF) {
	switch (c) {
	    case 'a':
		use_async = 0;
		break;
	    case 'r':
		rtr = TRUE;
		break;
	    case 'e':
		extd = TRUE;
		break;
	    case 'b':
		baud = atoi(optarg);
		break;
	    case 'l':
		load = atoi(optarg);
		doload = TRUE;
		break;
#ifdef USE_RT_SCHEDULING
	    case 'p':
	        {
	        struct sched_param mysched;
		    priority = atoi(optarg);
		    if (priority < 0 ) {
		      fprintf(stderr, "Priority < 0 not allowed\n");
		    }
		    if (priority > max_priority) {
		      fprintf(stderr, "Priority > %d not allowed\n",
		      					max_priority);
		    }
		    mysched.sched_priority =
		    		sched_get_priority_max(SCHED_RR) - 1;
		    ret = sched_setscheduler(0,SCHED_FIFO,&mysched);
		    if ( debug == TRUE ) {
			printf("sched_setscheduler() = %d\n", ret);
		    }
		    /* lock all currently and in future
			allocated memory blocks in physical ram */
		    ret = mlockall(MCL_CURRENT | MCL_FUTURE);
		    if ( debug == TRUE ) {
			printf("mlockall() = %d\n", ret);
		    }
		}
		break;
#endif
	    case 's':
		sleeptime = atoi(optarg);
		break;
	    case 'n':
		node = atoi(optarg);
		sprintf(device, "/dev/canp%d", node);
		break;
	    case 'D':
	        if (0 == strcmp(optarg, "stdout")) {
	            cstdout = TRUE;
	        } else {
		    sprintf(device, "/dev/%s", optarg);
		}
		break;
	    case 'd':
		debug = TRUE;
		break;
	    case 't':
		stresstest = TRUE;
		testtype   = atoi(optarg);
		break;
	    case 'T':
		test_count_max = atoi(optarg);
		break;
	    case 'V':
		printf("can_send V " VERSION ", " __DATE__ "\n");
		exit(0);
		break;
	    case 'R':
		canreset = TRUE;
		break;
	    case 'h':
	    default: 
	      usage(pname); 
	      exit(0);
	}
    }


    if ( argc - optind > 0 ) {
        /* at least one additional argument, the message id is given */
	message[0].id =  strtol(argv[optind++], NULL, 0);
    	memset(message[0].data, 0, 8);
	message[0].length = 0;
    }
    if ( argc - optind > 0 ) {
    	/* also data bytes areg given with the command */
	cnt = 0;
	while(optind != argc) {
	    message[0].data[cnt++] = strtol(argv[optind++], NULL, 0);
	}
	message[0].length = cnt;
    }
    if (rtr) {
	message[0].flags |= MSG_RTR;
    }
    if (extd) {
	message[0].flags |= MSG_EXT;
    }

    if ( debug == TRUE ) {

	printf("can_send V " VERSION ", " __DATE__ "\n");
	printf("(c) 1996-2003 port GmbH\n");
	printf(" using canmsg_t with %d bytes\n", (int)sizeof(canmsg_t));
	printf(" max process priority is \"-p %d\"\n", max_priority);
	if (stresstest) {
	    printf("should send one of the test sequences\n");
	} else {
	    printf("should send mess %ld with: %s", message[0].id,
		message[0].length > 0 ? ": " : "out data");
	    sbuf(message[0].data, message[0].length);
	}
    }

    sleeptime *= 1000;
    if ( debug == TRUE ) {
	printf("Sleeptime between transmit messages= %d us\n", sleeptime);
    }
    srand(node * 100);

    if (cstdout == FALSE) {
        /* really use CAN, open the device driver */
	if ( debug == TRUE ) {
	    printf("Open device %s\n", device);
	}
	can_fd = open(device, O_RDWR);
	if (can_fd == -1) {
	    fprintf(stderr, "open error %d;", errno);
	    perror(device);
	    exit(1);
	}
       
	printf("Open device %s can_fd %d \n", 
	       device,can_fd);

	//close(can_fd);
        //sleep(10);
	//printf("Test Close device %s can_fd %d \n", 
	//       device,can_fd);

	//can_fd = open(device, O_RDWR);
	if (can_fd == -1) {
	    fprintf(stderr, "open error %d;", errno);
	    perror(device);
	    exit(1);
	}
	can_fd2 = open(device, O_RDWR);
	if (can_fd2 == -1) {
	    fprintf(stderr, "open error %d;", errno);
	    perror(device);
	    exit(1);
	}
	can_fd3 = open(device, O_RDWR);
	if (can_fd3 == -1) {
	    fprintf(stderr, "open error %d;", errno);
	    perror(device);
	    exit(1);
	}


	printf("Open device %s can_fd %d can_fd2 %d can_fd3 %d\n", 
	       device,can_fd,can_fd2,can_fd3);


	if ( canreset == TRUE ) {
	    ret = can_reset();
	    if ( ret == -1 ) {
		perror("CAN Reset");
		exit(EXIT_FAILURE);
	    }
	    if ( debug == TRUE) {
		printf("Reset successfull\n");
	    }
	    exit(EXIT_SUCCESS);
	}
	
	if (baud > 0) {
	    if ( debug == TRUE ) {
		printf("change Bit-Rate to %d Kbit/s\n", baud);
	    }
	    set_bitrate(can_fd, baud);
	}
    } else {
	can_fd = 1;		/* use stdout */
    }

    if ( use_async > 0 ) {

      if ( can_fd > 1 ) {
	
	int fdflags;
	
	// PSW Set up Async
	act.sa_handler = can_async_handler;
	sigemptyset(&(act.sa_mask));
	act.sa_flags = 0;


	sigaction(SIGIO, &act, NULL);
	
	fcntl(can_fd,F_SETOWN,getpid());
	fdflags = fcntl(can_fd,F_GETFL);
	fcntl(can_fd,F_SETFL,fdflags | FASYNC );
	
      }
    }

    {
      // PSW get/set the masks
      int rbuffer = get_rbuffer(can_fd);
      int tbuffer = get_tbuffer(can_fd);
      
      //printf("PSW rbuffer = %x tbuffer = %x/n",rbuffer,tbuffer);
      tbuffer = 16; // PSW set a tx buffer mask
      rbuffer = 0; // PSW set a rx buffer mask ( we dont read )
      set_rtbuffer(can_fd,rbuffer,tbuffer);
      sleep(1);
    }


    if ( debug == TRUE ) {
	printf("opened %s succesful, got can_fd = %d\n", device, can_fd);
    }

    //set_rtbuffer(can_fd3,0,0);
    //set_rtbuffer(can_fd2,0,0);
    //set_rtr_reply(can_fd3); 

    if (rtr) {
	message[0].flags |= MSG_RTR;
	message[1].flags |= MSG_RTR;
	message[2].flags |= MSG_RTR;
	message[3].flags |= MSG_RTR;
    }

    /* the default can_send, simply send a message */
    /* no special test, send the normal message, (old behaviouur) */
    do {
        if ( debug == TRUE || 1) {
	     printf(" transmit message %ld\n", message[0].id ); 
	}
	//ret = write(can_fd, &message[0], 1);
	ret = write(can_fd, &message[0], 4);
        //ret = 4;
	if (ret == -1) {
	    /* int e = errno; */
	    perror("write error");
	    /* if ( e == ENOSPC) { */
		usleep(sleeptime); 
		continue;
	    /* } */
	} else if (ret == 0) {
	    printf("transmit timed out\n");
		usleep(sleeptime); 
		continue;
	} else {
	    printf("transmit one ret = %d\n",ret);
	}
    
	// try second write
	ret = write(can_fd, &message[0], 4);
        //ret = 4;
	if (ret == -1) {
	    /* int e = errno; */
	    perror("write 2 error");
	    /* if ( e == ENOSPC) { */
		usleep(sleeptime); 
		continue;
	    /* } */
	} else if (ret == 0) {
	    printf("write 2 transmit timed out\n");
		usleep(sleeptime); 
		continue;
	} else {
	    printf("transmit two ret = %d\n",ret);
	}
	if ( debug == TRUE ) {
	     getStat();
 	}
	// PSW try poll
	fds[0] = can_fd;
	fds[1] = can_fd2;
	//ret = can_poll(fds,2,10);  

        // PSW try many reads
	read_buff(can_fd,"can_fd");
	read_buff(can_fd2,"can_fd2");
	read_buff(can_fd3,"can_fd3");


	if (sleeptime > 0) usleep(sleeptime);
	message[0].id += increment;
	message[0].data[1]++;
    }
    while(sleeptime > 0);


    if (sleeptime == 0) {
        /* do not close while the controller is sending a message
         * sleep() then close()
         */
        usleep(50000);
        /* sleep(1); */
    }
    ret = close(can_fd);
    if (ret == -1) {
	fprintf(stderr, "close error %d;", errno);
	perror("");
	exit(1);
    }
    if ( debug == TRUE ) {
	printf("closed fd = %d succesful\n", can_fd);
    }
    return 0;
}


/* show buffer in hex */
void sbuf(unsigned char *s, int n)
{
int i;

    for(i = 0; i< n; i++) {
	fprintf(stdout, "%02x ", *s++);
    }
    putchar('\n');
}



void usage(char *s)
{
static char *usage_text  = "\
 ist optional id nicht angegeben wird 100 benutzt\n\
 sind keine Daten angegeben, werden max 8 ausgegeben\n\
 eingabe dezimal 100 or hex 0x64\n\
-r send message as rtr message.\n\
-e send message in extended message format.\n\
-l load try to reach this bus load, given in %\n\
-s n - sleeptime between messages in ms, if not specified send single message\n\
-d   - debug On\n\
       schaltet zusaetzlich Debugging im Treiber an/aus\n\
-b baudrate (Standard uses value of /proc/sys/Can/baud)\n\
-D dev use /dev/dev/{can0,can1,can2,can3} (real nodes, std: can1)\n\
"
"\
-t type \n\
   1 Stresstest fr Knoten, Sendet Bursts von kurzen rtr messages\n\
   2 sendet bursts von 5 Daten Messages, gleiche ID\n\
   3 sendet bursts von 5 Daten Messages, unterschiedliche ID\n\
     message 5 enthlt counter \n\
   4 as without this option, but incremnts CAN-ID with each message\n\
   10 sendet bursts von 9 Daten Messages, fr Comm. Verification\n\
   11 send -T number of messages as fast as possible, if transmit buffer\n\
      is full, sleep for -s ms time. time == 0:don't sleep, poll\n\
      after every message the messageid will increment\n\
   12 same as 11\n\
      but the message id is constant and the databytes will be increment\n\
-R   setzt nur CAN Controller zurck, danach exit()\n\
-T   Anzahl der Bursts, Abstand -s n (fr -t)\n\
-V   version\n\
\n\
";
    fprintf(stderr, "usage: %s options [id [ byte ..]]\n", s);
    fprintf(stderr, usage_text);


}

/* test1:
   - Data message 8 byte
   - RTR  Message 0 byte
   - RTR  Message 1 byte
   - RTR  Message 2 byte
*/
void test1(void)
{
long int test_count = 0;
canmsg_t tm[4];
int ret;

    tm[0].id = 101;
    tm[0].cob = 0;
    tm[0].length = 8;
    tm[0].flags = 0;
    if (extd) {
	tm[0].flags |= MSG_EXT;
    }
    tm[0].data[0] = 0x55;
    tm[0].data[1] = 0;
    tm[0].data[2] = 3;
    tm[0].data[3] = 4;
    tm[0].data[4] = 5;
    tm[0].data[5] = 6;
    tm[0].data[6] = 7;
    tm[0].data[7] = 0xaa;


    do {
        printf(" Psw tick %d \n",tm[0].data[1]);

	ret = write(can_fd, &tm[0], 1);
	if (ret == -1) {
	    perror("write error");
	    usleep(sleeptime); 
	    continue;
	} else if (ret == 0) {
	    printf("transmit timed out\n");
	    usleep(sleeptime); 
	    continue;
	} else {
	    if ( debug == TRUE ) {
		printf("transmitted %d\n", ret);
	    }
	}
	if (++test_count == test_count_max) {
	    break;
	}
        // PSW try a read
        ret = read(can_fd2, &tm[1], 1);
	if (ret == -1) {
	    perror("read error");
	    usleep(sleeptime); 
	    continue;
	} else if (ret == 0) {
	    printf("no read data\n");
	    usleep(sleeptime); 
	    continue;
	} else {
	    if ( 1 || debug == TRUE ) {
		printf("read %d\n", ret);
	    }
	}

	if (sleeptime > 0 )  {
	    usleep(sleeptime);
	}
	tm[0].data[1]++;
    }
    while ( sleeptime > 0 );
}

#endif
