/*   Receive2 polls all receive objects for a received message         */
/*   It is using the RECEIVE ioctl command                                */
/*   (c) 2003 port GmbH                                                */

 
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <can4linux.h>


#define STDDEV "/dev/can0"

#ifndef TRUE
# define TRUE  1
# define FALSE 0
#endif


int can_fd;
int sleeptime		 = 1000; 		/* in ms */
int debug                = FALSE;
int extd                 = FALSE;
int rtr                  = FALSE;

canmsg_t message;

void usage(char *s)
{
static char *usage_text  = "\
-d   - debug On\n\
       schaltet zusaetzlich Debugging im Treiber an/aus\n\
-s ms - sleep time ms between polling for a message\n\
-D dev use /dev/dev  - with dev: can[0123] (std: can0)\n\
\n\
";
    fprintf(stderr, "usage: %s [options] [id [ byte ..]]\n", s);
    fprintf(stderr, usage_text);
}

int main(int argc,char **argv)
{
int c;
extern char *optarg;
extern int optind, opterr, optopt;
int cnt;
char device[40] = STDDEV;
Receive_par_t RecTeil;

    while ((c = getopt(argc, argv, "dhs:D:")) != EOF) {
	switch (c) {
	    /* case 'b': */
		/* baud = atoi(optarg); */
		/* break; */
	    case 'D':
		sprintf(device, "/dev/%s", optarg);
		break;
	    case 'd':
		debug = TRUE;
		break;
	    case 's':
		sleeptime = atoi(optarg);
		break;
	    case 'h':
	    default: usage(argv[0]); exit(0);
	}
    }

    if(debug) printf("using CAN device %s\n", device);

    if(( can_fd = open(device, O_RDWR )) < 0 ) {
	fprintf(stderr,"Error opening CAN device %s\n", device);
        exit(1);
    }


    if ( debug == TRUE ) {
	printf("receive2 " __DATE__ "\n");
	printf("(c) 2003 port GmbH\n");
	printf(" using canmsg_t with %d bytes\n", (int)sizeof(canmsg_t));
    }

    RecTeil.Rx = &message;
    while(TRUE) {
	cnt = ioctl(can_fd, RECEIVE, &RecTeil);
	if(debug) {
	    printf("CAN RECEIVE returned with %d/0x%x\n", cnt, RecTeil.retval);
	}
	if(RecTeil.retval > 0) {
	    printf("Received id=%ld len=%d msg='%s' flags=0x%x \n",
		    RecTeil.Rx->id, RecTeil.Rx->length,
		    RecTeil.Rx->data ,RecTeil.Rx->flags );
	}
	usleep(sleeptime * 1000);
    }

    close(can_fd);
    return 0;
}

